(*
(C) Copyright Andreas Viktor Hess, DTU, 2018

All Rights Reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:

- Redistributions of source code must retain the above copyright
  notice, this list of conditions and the following disclaimer.

- Redistributions in binary form must reproduce the above copyright
  notice, this list of conditions and the following disclaimer in the
  documentation and/or other materials provided with the distribution.

- Neither the name of the copyright holder nor the names of its
  contributors may be used to endorse or promote products
  derived from this software without specific prior written
  permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
"AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*)

(*  Title:      Stateful_Compositionality.thy
    Author:     Andreas Viktor Hess, DTU
*)

theory Stateful_Compositionality
imports Stateful_Typing Parallel_Compositionality
begin

abbreviation dbproj where "dbproj l D \<equiv> filter (\<lambda>d. fst d = l) D"

section {* Small lemmas *}
lemma labeled_list_insert_eq_cases:
  "d \<notin> set (unlabel D) \<Longrightarrow> List.insert d (unlabel D) = unlabel (List.insert (i,d) D)"
  "(i,d) \<in> set D \<Longrightarrow> List.insert d (unlabel D) = unlabel (List.insert (i,d) D)"
unfolding unlabel_def
by (metis (no_types, hide_lams) List.insert_def image_eqI list.simps(9) set_map snd_conv,
    metis in_set_insert set_zip_rightD zip_map_fst_snd)

lemma labeled_list_insert_eq_ex_cases:
  "List.insert d (unlabel D) = unlabel (List.insert (i,d) D) \<or>
  (\<exists>j. (j,d) \<in> set D \<and> List.insert d (unlabel D) = unlabel (List.insert (j,d) D))"
using labeled_list_insert_eq_cases unfolding unlabel_def
by (metis in_set_impl_in_set_zip2 length_map zip_map_fst_snd)

lemma proj_set_subset:
  "set (proj n A) \<subseteq> set A"
unfolding proj_def by auto

lemma proj_proj_set_subset:
  "set (proj n (proj m A)) \<subseteq> set (proj n A)"
  "set (proj n (proj m A)) \<subseteq> set (proj m A)"
  "set (proj_unl n (proj m A)) \<subseteq> set (proj_unl n A)"
  "set (proj_unl n (proj m A)) \<subseteq> set (proj_unl m A)"
unfolding unlabel_def proj_def by auto

lemma proj_in_set_iff:
  "(ln i, d) \<in> set (proj i D) \<longleftrightarrow> (ln i, d) \<in> set D"
  "(\<star>, d) \<in> set (proj i D) \<longleftrightarrow> (\<star>, d) \<in> set D"
unfolding proj_def by auto

lemma proj_list_insert:
  "proj i (List.insert (ln i,d) D) = List.insert (ln i,d) (proj i D)"
  "proj i (List.insert (\<star>,d) D) = List.insert (\<star>,d) (proj i D)"
  "i \<noteq> j \<Longrightarrow> proj i (List.insert (ln j,d) D) = proj i D"
unfolding List.insert_def proj_def by auto

lemma proj_filter: "proj i [d\<leftarrow>D. d \<notin> set Di] = [d\<leftarrow>proj i D. d \<notin> set Di]"
by (simp_all add: proj_def conj_commute)

lemma proj_list_Cons:
  "proj i ((ln i,d)#D) = (ln i,d)#proj i D"
  "proj i ((\<star>,d)#D) = (\<star>,d)#proj i D"
  "i \<noteq> j \<Longrightarrow> proj i ((ln j,d)#D) = proj i D"
unfolding List.insert_def proj_def by auto

lemma proj_dbproj: 
  "dbproj (ln i) (proj i D) = dbproj (ln i) D"
  "dbproj \<star> (proj i D) = dbproj \<star> D"
  "i \<noteq> j \<Longrightarrow> dbproj (ln j) (proj i D) = []"
unfolding proj_def by (induct D) auto

lemma dbproj_Cons:
  "dbproj i ((i,d)#D) = (i,d)#dbproj i D"
  "i \<noteq> j \<Longrightarrow> dbproj j ((i,d)#D) = dbproj j D"
by auto

lemma dbproj_subseq: 
  assumes "Di \<in> set (subseqs (dbproj k D))"
  shows "dbproj k Di = Di" (is ?A)
  and "i \<noteq> k \<Longrightarrow> dbproj i Di = []" (is "i \<noteq> k \<Longrightarrow> ?B")
proof -
  have *: "set Di \<subseteq> set (dbproj k D)" using subseqs_powset[of "dbproj k D"] assms by auto
  thus ?A by (metis filter_True filter_set member_filter subsetCE)

  have "\<And>j d. (j,d) \<in> set Di \<Longrightarrow> j = k" using * by auto
  moreover have "\<And>j d. (j,d) \<in> set (dbproj i Di) \<Longrightarrow> j = i" by auto
  moreover have "\<And>j d. (j,d) \<in> set (dbproj i Di) \<Longrightarrow> (j,d) \<in> set Di" by auto
  ultimately show "i \<noteq> k \<Longrightarrow> ?B" by (metis set_empty subrelI subset_empty)
qed

lemma dbproj_subseq_subset:
  assumes "Di \<in> set (subseqs (dbproj i D))"
  shows "set Di \<subseteq> set D"
by (metis Pow_iff assms filter_set image_eqI member_filter subseqs_powset subsetCE subsetI)

lemma dbproj_subseq_in_subseqs:
  assumes "Di \<in> set (subseqs (dbproj i D))"
  shows "Di \<in> set (subseqs D)"
using assms in_set_subseqs subseq_filter_left subseq_order.dual_order.trans by blast

lemma proj_subseq:
  assumes "Di \<in> set (subseqs (dbproj (ln j) D))" "j \<noteq> i"
  shows "[d\<leftarrow>proj i D. d \<notin> set Di] = proj i D"
proof -
  have "set Di \<subseteq> set (dbproj (ln j) D)" using subseqs_powset[of "dbproj (ln j) D"] assms by auto
  hence "\<And>k d. (k,d) \<in> set Di \<Longrightarrow> k = ln j" by auto
  moreover have "\<And>k d. (k,d) \<in> set (proj i D) \<Longrightarrow> k \<noteq> ln j"
    using assms(2) unfolding proj_def by auto
  ultimately have "\<And>d. d \<in> set (proj i D) \<Longrightarrow> d \<notin> set Di" by auto
  thus ?thesis by simp
qed

lemma map_subseqs: "map (map f) (subseqs A) = subseqs (map f A)"
proof (induct A)
  case (Cons a A)
  have "map (Cons (f a)) (map (map f) (subseqs A)) = map (map f) (map (Cons a) (subseqs A))"
    by (induct "subseqs A") auto
  thus ?case by (simp add: Let_def Cons)
qed simp

lemma unlabel_subseqsD:
  assumes "A \<in> set (subseqs (unlabel B))"
  shows "\<exists>C \<in> set (subseqs B). unlabel C = A"
using assms map_subseqs unfolding unlabel_def by (metis imageE set_map) 

lemma suffix_Cons': "suffix [b] (a#A) \<Longrightarrow> suffix [b] A \<or> (a = b \<and> A = [])"
using suffix_Cons[of "[b]"] by auto

lemma prefix_Cons': "prefix (a#A) (a#B) \<Longrightarrow> prefix A B"
by simp

lemma map_snd_list_insert_distrib:
  assumes "\<forall>(i,p) \<in> insert d (set D). \<forall>(i',p') \<in> insert d (set D). p = p' \<longrightarrow> i = i'"
  shows "map snd (List.insert d D) = List.insert (snd d) (map snd D)"
using assms
proof (induction D rule: List.rev_induct)
  case (snoc a D)
  hence IH: "map snd (List.insert d D) = List.insert (snd d) (map snd D)" by fastforce

  obtain ia pa where a: "a = (ia,pa)" by (metis surj_pair)
  obtain id pd where d: "d = (id,pd)" by (metis surj_pair)

  have "(id,pd) \<in> insert d (set (a#D))" "(ia,pa) \<in> insert d (set (a#D))" using a d by auto
  hence *: "pa = pd \<Longrightarrow> ia = id" using snoc.prems by auto

  show ?case
  proof (cases "pd = pa")
    case True
    hence "a = d" using * a d by auto
    thus ?thesis using IH by simp
  next
    case False
    hence "a \<noteq> d" using a d by simp
    have "List.insert d (D@[a]) = (List.insert d D)@[a]"
    proof -
      have 1: "insert a (set D) = set (D@[a])" by simp
      have 2: "d \<notin> insert a (set D) \<or> d \<in> set D" using \<open>a \<noteq> d\<close> by blast
      show ?thesis using 1 2 by (metis (no_types) List.insert_def append_Cons insertCI)
    qed
    thus ?thesis using IH a d False by (auto simp add: List.insert_def)
  qed
qed simp

lemma unlabel_filter_eq:
  assumes "\<forall>(j, p) \<in> set A \<union> B. \<forall>(k, q) \<in> set A \<union> B. p = q \<longrightarrow> j = k" (is "?P (set A)")
  shows "[d\<leftarrow>unlabel A. d \<notin> snd ` B] = unlabel [d\<leftarrow>A. d \<notin> B]"
using assms unfolding unlabel_def
proof (induction A)
  case (Cons a A)
  have "set A \<subseteq> set (a#A)" "{a} \<subseteq> set (a#A)" by auto
  hence *: "?P (set A)" "?P {a}" using Cons.prems by fast+
  hence IH: "[d\<leftarrow>map snd A . d \<notin> snd ` B] = map snd [d\<leftarrow>A . d \<notin> B]" using Cons.IH by auto

  { assume "snd a \<in> snd ` B"
    then obtain b where b: "b \<in> B" "snd a = snd b" by moura
    hence "fst a = fst b" using *(2) by auto
    hence "a \<in> B" using b by (metis surjective_pairing)  
  } hence **: "a \<notin> B \<Longrightarrow> snd a \<notin> snd ` B" by metis

  show ?case by (cases "a \<in> B") (simp add: ** IH)+ 
qed simp

lemma subseqs_Cons:
  assumes "A \<in> set (subseqs B)"
  shows "A \<in> set (subseqs (b#B))"
by (metis assms Un_iff set_append subseqs.simps(2))

lemma subseqs_subset:
  assumes "Di \<in> set (subseqs D)"
  shows "set Di \<subseteq> set D"
using assms by (metis Pow_iff image_eqI subseqs_powset)

lemma subseqs_mem_dbproj:
  assumes "Di \<in> set (subseqs D)" "list_all (\<lambda>d. fst d = i) Di"
  shows "Di \<in> set (subseqs (dbproj i D))"
using assms
proof (induction D arbitrary: Di)
  case (Cons di D)
  obtain d j where di: "di = (j,d)" by (metis surj_pair)
  show ?case
  proof (cases "Di \<in> set (subseqs D)")
    case True
    hence "Di \<in> set (subseqs (dbproj i D))" using Cons.IH Cons.prems by auto
    thus ?thesis using subseqs_Cons by auto
  next
    case False
    then obtain Di' where Di': "Di = di#Di'" using Cons.prems(1)
      by (metis (mono_tags, lifting) Un_iff imageE set_append set_map subseqs.simps(2)) 
    hence "Di' \<in> set (subseqs D)" using Cons.prems(1) False
      by (metis (no_types, lifting) UnE imageE list.inject set_append set_map subseqs.simps(2)) 
    hence "Di' \<in> set (subseqs (dbproj i D))" using Cons.IH Cons.prems Di' by auto
    moreover have "i = j" using Di' di Cons.prems(2) by auto
    hence "dbproj i (di#D) = di#dbproj i D" by (simp add: di)
    ultimately show ?thesis using Di'
      by (metis (no_types, lifting) UnCI image_eqI set_append set_map subseqs.simps(2)) 
  qed
qed simp

lemma (in stateful_typed_model) labeled_sat_ineq_lift:
  assumes "\<lbrakk>M; map (\<lambda>d. \<nexists>X\<langle>to_pair (t,s) \<doteq> to_pair (snd d)\<rangle>) [d\<leftarrow>dbproj i D. d \<notin> set Di]\<rbrakk>\<^sub>d \<I>"
    (is "?R1 D")
  and "\<forall>(j,p) \<in> {(i,t,s)} \<union> set D \<union> set Di. \<forall>(k,q) \<in> {(i,t,s)} \<union> set D \<union> set Di.
          (\<exists>\<delta>. Unifier \<delta> (to_pair p) (to_pair q)) \<longrightarrow> j = k" (is "?R2 D")
  shows "\<lbrakk>M; map (\<lambda>d. \<nexists>X\<langle>to_pair (t,s) \<doteq> to_pair (snd d)\<rangle>) [d\<leftarrow>D. d \<notin> set Di]\<rbrakk>\<^sub>d \<I>"
using assms
proof (induction D)
  case (Cons dl D)
  obtain d l where dl: "dl = (l,d)" by (metis surj_pair)

  have 1: "?R1 D"
  proof (cases "i = l")
    case True thus ?thesis using Cons.prems(1) dl by (cases "dl \<in> set Di") auto
  next
    case False thus ?thesis using Cons.prems(1) dl by auto
  qed

  have "set D \<subseteq> set (dl#D)" by auto
  hence 2: "?R2 D" using Cons.prems(2) by blast

  have "i \<noteq> l \<or> dl \<in> set Di \<or> \<lbrakk>M; [\<nexists>X\<langle>to_pair (t, s) \<doteq> to_pair (snd dl)\<rangle>]\<rbrakk>\<^sub>d \<I>"
    using Cons.prems(1) dl by (auto simp add: ineq_model_def)
  moreover have "\<exists>\<delta>. Unifier \<delta> (to_pair (t,s)) (to_pair d) \<Longrightarrow> i = l"
    using Cons.prems(2) dl by force
  ultimately have 3: "dl \<in> set Di \<or> \<lbrakk>M; [\<nexists>X\<langle>to_pair (t, s) \<doteq> to_pair (snd dl)\<rangle>]\<rbrakk>\<^sub>d \<I>"
    using strand_sem_not_unif_is_sat_ineq[of "to_pair (t,s)" "to_pair d"] dl by fastforce

  show ?case using Cons.IH[OF 1 2] 3 dl by auto
qed simp

lemma (in stateful_typed_model) labeled_sat_ineq_lift':
  assumes "\<lbrakk>M; map (\<lambda>d. \<nexists>X\<langle>to_pair (t,s) \<doteq> to_pair (snd d)\<rangle>) (dbproj i D)\<rbrakk>\<^sub>d \<I>"
    (is "?R1 D")
  and "\<forall>(j,p) \<in> {(i,t,s)} \<union> set D. \<forall>(k,q) \<in> {(i,t,s)} \<union> set D.
          (\<exists>\<delta>. Unifier \<delta> (to_pair p) (to_pair q)) \<longrightarrow> j = k" (is "?R2 D")
  shows "\<lbrakk>M; map (\<lambda>d. \<nexists>X\<langle>to_pair (t,s) \<doteq> to_pair (snd d)\<rangle>) D\<rbrakk>\<^sub>d \<I>"
using assms
proof (induction D)
  case (Cons dl D)
  obtain d l where dl: "dl = (l,d)" by (metis surj_pair)

  have 1: "?R1 D" using Cons.prems(1) dl by (cases "i = l") auto

  have "set D \<subseteq> set (dl#D)" by auto
  hence 2: "?R2 D" using Cons.prems(2) by blast

  have "i \<noteq> l \<or> \<lbrakk>M; [\<nexists>X\<langle>to_pair (t, s) \<doteq> to_pair (snd dl)\<rangle>]\<rbrakk>\<^sub>d \<I>"
    using Cons.prems(1) dl by auto
  moreover have "\<exists>\<delta>. Unifier \<delta> (to_pair (t,s)) (to_pair d) \<Longrightarrow> i = l"
    using Cons.prems(2) dl by force
  ultimately have 3: "\<lbrakk>M; [\<nexists>X\<langle>to_pair (t, s) \<doteq> to_pair (snd dl)\<rangle>]\<rbrakk>\<^sub>d \<I>"
    using strand_sem_not_unif_is_sat_ineq[of "to_pair (t,s)" "to_pair d"] dl by fastforce

  show ?case using Cons.IH[OF 1 2] 3 dl by auto
qed simp

lemma (in stateful_typed_model) labeled_sat_ineq_dbproj:
  assumes "\<lbrakk>M; map (\<lambda>d. \<nexists>X\<langle>to_pair (t,s) \<doteq> to_pair (snd d)\<rangle>) [d\<leftarrow>D. d \<notin> set Di]\<rbrakk>\<^sub>d \<I>"
    (is "?P D")
  shows "\<lbrakk>M; map (\<lambda>d. \<nexists>X\<langle>to_pair (t,s) \<doteq> to_pair (snd d)\<rangle>) [d\<leftarrow>dbproj i D. d \<notin> set Di]\<rbrakk>\<^sub>d \<I>"
    (is "?Q D")
using assms
proof (induction D)
  case (Cons di D)
  obtain d j where di: "di = (j,d)" by (metis surj_pair)

  have "?P D" using Cons.prems by (cases "di \<in> set Di") auto
  hence IH: "?Q D" by (metis Cons.IH)

  show ?case using di IH
  proof (cases "i = j \<and> di \<notin> set Di")
    case True
    have 1: "\<lbrakk>M; [\<nexists>X\<langle>to_pair (t, s) \<doteq> to_pair (snd di)\<rangle>]\<rbrakk>\<^sub>d \<I>" using Cons.prems True by auto
    have 2: "dbproj i (di#D) = di#dbproj i D" using True dbproj_Cons(1) di by auto
    show ?thesis using 1 2 IH by auto
  qed auto
qed simp

lemma (in stateful_typed_model) labeled_sat_ineq_dbproj':
  assumes "\<lbrakk>M; map (\<lambda>d. \<nexists>X\<langle>to_pair (t,s) \<doteq> to_pair (snd d)\<rangle>) D\<rbrakk>\<^sub>d \<I>"
    (is "?P D")
  shows "\<lbrakk>M; map (\<lambda>d. \<nexists>X\<langle>to_pair (t,s) \<doteq> to_pair (snd d)\<rangle>) (dbproj i D)\<rbrakk>\<^sub>d \<I>"
    (is "?Q D")
using assms
proof (induction D)
  case (Cons di D)
  obtain d j where di: "di = (j,d)" by (metis surj_pair)

  have "?P D" using Cons.prems by auto
  hence IH: "?Q D" by (metis Cons.IH)

  show ?case using di IH
  proof (cases "i = j")
    case True
    have 1: "\<lbrakk>M; [\<nexists>X\<langle>to_pair (t, s) \<doteq> to_pair (snd di)\<rangle>]\<rbrakk>\<^sub>d \<I>" using Cons.prems True by auto
    have 2: "dbproj i (di#D) = di#dbproj i D" using True dbproj_Cons(1) di by auto
    show ?thesis using 1 2 IH by auto
  qed auto
qed simp

lemma (in stateful_typed_model) labeled_sat_eqs_list_all:
  assumes "\<forall>(j, p) \<in> {(i,t,s)} \<union> set D. \<forall>(k,q) \<in> {(i,t,s)} \<union> set D.
              (\<exists>\<delta>. Unifier \<delta> (to_pair p) (to_pair q)) \<longrightarrow> j = k" (is "?P D")
  and "\<lbrakk>M; map (\<lambda>d. \<langle>to_pair (t,s) \<doteq> to_pair (snd d)\<rangle>) D\<rbrakk>\<^sub>d \<I>" (is "?Q D")
  shows "list_all (\<lambda>d. fst d = i) D"
using assms
proof (induction D rule: List.rev_induct)
  case (snoc di D)
  obtain d j where di: "di = (j,d)" by (metis surj_pair)
  have "to_pair (t,s) \<cdot> \<I> = to_pair d \<cdot> \<I>" using di snoc.prems(2) by auto
  hence "\<exists>\<delta>. Unifier \<delta> (to_pair (t,s)) (to_pair d)" by auto
  hence 1: "i = j" using snoc.prems(1) di by fastforce
  
  have "set D \<subseteq> set (D@[di])" by auto
  hence 2: "?P D" using snoc.prems(1) by blast

  have 3: "?Q D" using snoc.prems(2) by auto
  
  show ?case using di 1 snoc.IH[OF 2 3] by simp
qed simp

lemma (in stateful_typed_model) labeled_sat_eqs_subseqs:
  assumes "Di \<in> set (subseqs D)"
  and "\<forall>(j, p) \<in> {(i,t,s)} \<union> set D. \<forall>(k, q) \<in> {(i,t,s)} \<union> set D.
          (\<exists>\<delta>. Unifier \<delta> (to_pair p) (to_pair q)) \<longrightarrow> j = k" (is "?P D")
  and "\<lbrakk>M; map (\<lambda>d. \<langle>to_pair (t, s) \<doteq> to_pair (snd d)\<rangle>) Di\<rbrakk>\<^sub>d \<I>"
  shows "Di \<in> set (subseqs (dbproj i D))"
proof -
  have "set Di \<subseteq> set D" by (rule subseqs_subset[OF assms(1)])
  hence "?P Di" using assms(2) by blast
  thus ?thesis using labeled_sat_eqs_list_all[OF _ assms(3)] subseqs_mem_dbproj[OF assms(1)] by simp
qed


section {* Locale setup and definitions *}

locale labeled_stateful_typed_model =
  stateful_typed_model arity public Ana \<Gamma> pair
+ labeled_typed_model arity public Ana \<Gamma> label_witness1 label_witness2
  for arity::"'fun \<Rightarrow> nat"
  and public::"'fun \<Rightarrow> bool"
  and Ana::"('fun,'var) term \<Rightarrow> (('fun,'var) terms \<times> ('fun,'var) terms)"
  and \<Gamma>::"('fun,'var) term \<Rightarrow> ('fun,'atom::finite) term_type"
  and pair::"'fun"
  and label_witness1::"'lbl"
  and label_witness2::"'lbl"
begin

type_synonym ('a,'b,'c) labeled_stateful_strand_step = "'c strand_label \<times> ('a,'b) stateful_strand_step"
type_synonym ('a,'b,'c) labeled_stateful_strand = "('a,'b,'c) labeled_stateful_strand_step list"

fun setops\<^sub>l\<^sub>s\<^sub>s\<^sub>t where
  "setops\<^sub>l\<^sub>s\<^sub>s\<^sub>t [] = {}"
| "setops\<^sub>l\<^sub>s\<^sub>s\<^sub>t ((i,Insert t s)#S) = insert (i,t,s) (setops\<^sub>l\<^sub>s\<^sub>s\<^sub>t S)"
| "setops\<^sub>l\<^sub>s\<^sub>s\<^sub>t ((i,Delete t s)#S) = insert (i,t,s) (setops\<^sub>l\<^sub>s\<^sub>s\<^sub>t S)"
| "setops\<^sub>l\<^sub>s\<^sub>s\<^sub>t ((i,InSet t s)#S) = insert (i,t,s) (setops\<^sub>l\<^sub>s\<^sub>s\<^sub>t S)"
| "setops\<^sub>l\<^sub>s\<^sub>s\<^sub>t ((i,NotInSet _ t s)#S) = insert (i,t,s) (setops\<^sub>l\<^sub>s\<^sub>s\<^sub>t S)"
| "setops\<^sub>l\<^sub>s\<^sub>s\<^sub>t (_#S) = setops\<^sub>l\<^sub>s\<^sub>s\<^sub>t S"

definition setops\<^sub>l\<^sub>d\<^sub>b where
  "setops\<^sub>l\<^sub>d\<^sub>b D = set (map (\<lambda>(i,p). (i,to_pair p)) D)"

definition par_comp\<^sub>l\<^sub>s\<^sub>s\<^sub>t where
  "par_comp\<^sub>l\<^sub>s\<^sub>s\<^sub>t (\<A>::('fun,'var,'lbl) labeled_stateful_strand) (Secrets::('fun,'var) terms) \<equiv> 
    (\<forall>l1 l2. l1 \<noteq> l2 \<longrightarrow>
              GSMP_disjoint (trms\<^sub>s\<^sub>s\<^sub>t (proj_unl l1 \<A>) \<union> setops\<^sub>s\<^sub>s\<^sub>t (proj_unl l1 \<A>))
                            (trms\<^sub>s\<^sub>s\<^sub>t (proj_unl l2 \<A>) \<union> setops\<^sub>s\<^sub>s\<^sub>t (proj_unl l2 \<A>)) Secrets) \<and>
    (\<forall>t. (\<star>, Send t) \<notin> set \<A>) \<and> ground Secrets \<and>
    (\<forall>s \<in> Secrets. \<forall>s' \<in> subterms s. {} \<turnstile>\<^sub>c s' \<or> s' \<in> Secrets) \<and>
    (\<forall>(i,p) \<in> setops\<^sub>l\<^sub>s\<^sub>s\<^sub>t \<A>. \<forall>(j,q) \<in> setops\<^sub>l\<^sub>s\<^sub>s\<^sub>t \<A>.
      (\<exists>\<delta>. Unifier \<delta> (to_pair p) (to_pair q)) \<longrightarrow> i = j)"

definition declassified\<^sub>l\<^sub>s\<^sub>s\<^sub>t where
  "declassified\<^sub>l\<^sub>s\<^sub>s\<^sub>t \<A> \<I> \<equiv> {t. (\<star>, stateful_strand_step.Receive t) \<in> set \<A>} \<cdot>\<^sub>s\<^sub>e\<^sub>t \<I>"

definition strand_leaks\<^sub>l\<^sub>s\<^sub>s\<^sub>t ("_ leaks _ under _") where
  "(\<A>::('fun,'var,'lbl) labeled_stateful_strand) leaks Secrets under \<I> \<equiv>
    (\<exists>t \<in> Secrets - declassified\<^sub>l\<^sub>s\<^sub>s\<^sub>t \<A> \<I>. \<exists>n. \<I> \<Turnstile>\<^sub>s (proj_unl n \<A>@[Send t]))"

definition typing_cond\<^sub>s\<^sub>s\<^sub>t where
  "typing_cond\<^sub>s\<^sub>s\<^sub>t \<A> \<equiv> wf\<^sub>s\<^sub>s\<^sub>t \<A> \<and> wf\<^sub>t\<^sub>r\<^sub>m\<^sub>s (trms\<^sub>s\<^sub>s\<^sub>t \<A>) \<and> tfr\<^sub>s\<^sub>s\<^sub>t \<A>"

type_synonym ('a,'b,'c) labeleddbstate = "('c strand_label \<times> (('a,'b) term \<times> ('a,'b) term)) set"
type_synonym ('a,'b,'c) labeleddbstatelist = "('c strand_label \<times> (('a,'b) term \<times> ('a,'b) term)) list"

fun tr\<^sub>p\<^sub>c::
  "('fun,'var,'lbl) labeled_stateful_strand \<Rightarrow> ('fun,'var,'lbl) labeleddbstatelist
   \<Rightarrow> ('fun,'var,'lbl) labeled_strand list"
where
  "tr\<^sub>p\<^sub>c [] D = [[]]"
| "tr\<^sub>p\<^sub>c ((i,Send t)#A) D = map ((#) (i,send\<langle>t\<rangle>)) (tr\<^sub>p\<^sub>c A D)"
| "tr\<^sub>p\<^sub>c ((i,Receive t)#A) D = map ((#) (i,receive\<langle>t\<rangle>)) (tr\<^sub>p\<^sub>c A D)"
| "tr\<^sub>p\<^sub>c ((i,Equality t t')#A) D = map ((#) (i,\<langle>t \<doteq> t'\<rangle>)) (tr\<^sub>p\<^sub>c A D)"
| "tr\<^sub>p\<^sub>c ((i,Inequality X t t')#A) D = map ((#) (i,\<nexists>X\<langle>t \<doteq> t'\<rangle>)) (tr\<^sub>p\<^sub>c A D)"
| "tr\<^sub>p\<^sub>c ((i,Insert t s)#A) D = tr\<^sub>p\<^sub>c A (List.insert (i,(t,s)) D)"
| "tr\<^sub>p\<^sub>c ((i,Delete t s)#A) D = (
    concat (map (\<lambda>Di. map (\<lambda>B. (map (\<lambda>d. (i,\<langle>to_pair (t,s) \<doteq> to_pair (snd d)\<rangle>)) Di)@
                               (map (\<lambda>d. (i,\<nexists>[]\<langle>to_pair (t,s) \<doteq> to_pair (snd d)\<rangle>))
                                    [d\<leftarrow>dbproj i D. d \<notin> set Di])@B)
                          (tr\<^sub>p\<^sub>c A [d\<leftarrow>D. d \<notin> set Di]))
                (subseqs (dbproj i D))))"
| "tr\<^sub>p\<^sub>c ((i,InSet t s)#A) D =
    concat (map (\<lambda>B. map (\<lambda>d. (i,\<langle>to_pair (t,s) \<doteq> to_pair (snd d)\<rangle>)#B) (dbproj i D)) (tr\<^sub>p\<^sub>c A D))"
| "tr\<^sub>p\<^sub>c ((i,NotInSet X t s)#A) D =
    map (\<lambda>B. (map (\<lambda>d. (i,\<nexists>X\<langle>to_pair (t,s) \<doteq> to_pair (snd d)\<rangle>)) (dbproj i D))@B) (tr\<^sub>p\<^sub>c A D)"


fun tr'\<^sub>p\<^sub>c::
  "('fun,'var,'lbl) labeled_stateful_strand \<Rightarrow> ('fun,'var,'lbl) labeleddbstatelist
   \<Rightarrow> ('fun,'var,'lbl) labeled_strand list"
where
  "tr'\<^sub>p\<^sub>c [] D = [[]]"
| "tr'\<^sub>p\<^sub>c ((i,Send t)#A) D = map ((#) (i,send\<langle>t\<rangle>)) (tr'\<^sub>p\<^sub>c A D)"
| "tr'\<^sub>p\<^sub>c ((i,Receive t)#A) D = map ((#) (i,receive\<langle>t\<rangle>)) (tr'\<^sub>p\<^sub>c A D)"
| "tr'\<^sub>p\<^sub>c ((i,Equality t t')#A) D = map ((#) (i,\<langle>t \<doteq> t'\<rangle>)) (tr'\<^sub>p\<^sub>c A D)"
| "tr'\<^sub>p\<^sub>c ((i,Inequality X t t')#A) D = map ((#) (i,\<nexists>X\<langle>t \<doteq> t'\<rangle>)) (tr'\<^sub>p\<^sub>c A D)"
| "tr'\<^sub>p\<^sub>c ((i,Insert t s)#A) D = tr'\<^sub>p\<^sub>c A (List.insert (i,(t,s)) D)"
| "tr'\<^sub>p\<^sub>c ((i,Delete t s)#A) D = (
    concat (map (\<lambda>Di. map (\<lambda>B. (map (\<lambda>d. (fst d,\<langle>to_pair (t,s) \<doteq> to_pair (snd d)\<rangle>)) Di)@
                               (map (\<lambda>d. (fst d,\<nexists>[]\<langle>to_pair (t,s) \<doteq> to_pair (snd d)\<rangle>))
                                    [d\<leftarrow>D. d \<notin> set Di])@B)
                          (tr'\<^sub>p\<^sub>c A [d\<leftarrow>D. d \<notin> set Di]))
                (subseqs D)))"
| "tr'\<^sub>p\<^sub>c ((i,InSet t s)#A) D =
    concat (map (\<lambda>B. map (\<lambda>d. (fst d,\<langle>to_pair (t,s) \<doteq> to_pair (snd d)\<rangle>)#B) D) (tr'\<^sub>p\<^sub>c A D))"
| "tr'\<^sub>p\<^sub>c ((i,NotInSet X t s)#A) D =
    map (\<lambda>B. (map (\<lambda>d. (fst d,\<nexists>X\<langle>to_pair (t,s) \<doteq> to_pair (snd d)\<rangle>)) D)@B) (tr'\<^sub>p\<^sub>c A D)"


section {* Lemmata: Small stuff *}

lemma sst_vars_proj_subset:
  "fv\<^sub>s\<^sub>s\<^sub>t (proj_unl n A) \<subseteq> fv\<^sub>s\<^sub>s\<^sub>t (unlabel A)"
  "bvars\<^sub>s\<^sub>s\<^sub>t (proj_unl n A) \<subseteq> bvars\<^sub>s\<^sub>s\<^sub>t (unlabel A)"
unfolding unlabel_def proj_def by auto

lemma sst_vars_append_subset:
  "fv\<^sub>s\<^sub>s\<^sub>t A \<subseteq> fv\<^sub>s\<^sub>s\<^sub>t (A@B)" "bvars\<^sub>s\<^sub>s\<^sub>t A \<subseteq> bvars\<^sub>s\<^sub>s\<^sub>t (A@B)"
  "fv\<^sub>s\<^sub>s\<^sub>t B \<subseteq> fv\<^sub>s\<^sub>s\<^sub>t (A@B)" "bvars\<^sub>s\<^sub>s\<^sub>t B \<subseteq> bvars\<^sub>s\<^sub>s\<^sub>t (A@B)"
unfolding unlabel_def proj_def by auto

lemma trms\<^sub>s\<^sub>s\<^sub>t_mono:
  "set M \<subseteq> set N \<Longrightarrow> trms\<^sub>s\<^sub>s\<^sub>t M \<subseteq> trms\<^sub>s\<^sub>s\<^sub>t N"
by auto

lemma trms\<^sub>s\<^sub>s\<^sub>t_proj_subset:
  "trms\<^sub>s\<^sub>s\<^sub>t (proj_unl n A) \<subseteq> trms\<^sub>s\<^sub>s\<^sub>t (unlabel A)" (is ?A)
  "trms\<^sub>s\<^sub>s\<^sub>t (proj_unl m (proj n A)) \<subseteq> trms\<^sub>s\<^sub>s\<^sub>t (proj_unl n A)" (is ?B)
  "trms\<^sub>s\<^sub>s\<^sub>t (proj_unl m (proj n A)) \<subseteq> trms\<^sub>s\<^sub>s\<^sub>t (proj_unl m A)" (is ?C)
proof -
  show ?A unfolding unlabel_def proj_def by auto
  show ?B using trms\<^sub>s\<^sub>s\<^sub>t_mono[OF proj_proj_set_subset(4)] by metis
  show ?C using trms\<^sub>s\<^sub>s\<^sub>t_mono[OF proj_proj_set_subset(3)] by metis
qed

lemma trms\<^sub>s\<^sub>s\<^sub>t_prefix_subset:
  "trms\<^sub>s\<^sub>s\<^sub>t (unlabel A) \<subseteq> trms\<^sub>s\<^sub>s\<^sub>t (unlabel (A@B))" (is ?A)
  "trms\<^sub>s\<^sub>s\<^sub>t (proj_unl n A) \<subseteq> trms\<^sub>s\<^sub>s\<^sub>t (proj_unl n (A@B))" (is ?B)
using trms\<^sub>s\<^sub>s\<^sub>t_mono[of "proj_unl n A" "proj_unl n (A@B)"]
unfolding unlabel_def proj_def by auto

lemma trms\<^sub>s\<^sub>s\<^sub>t_suffix_subset:
  "trms\<^sub>s\<^sub>s\<^sub>t (unlabel B) \<subseteq> trms\<^sub>s\<^sub>s\<^sub>t (unlabel (A@B))"
  "trms\<^sub>s\<^sub>s\<^sub>t (proj_unl n B) \<subseteq> trms\<^sub>s\<^sub>s\<^sub>t (proj_unl n (A@B))"
using trms\<^sub>s\<^sub>s\<^sub>t_mono[of "proj_unl n B" "proj_unl n (A@B)"]
unfolding unlabel_def proj_def by auto

lemma setops\<^sub>s\<^sub>s\<^sub>t_proj_subset:
  "setops\<^sub>s\<^sub>s\<^sub>t (proj_unl n A) \<subseteq> setops\<^sub>s\<^sub>s\<^sub>t (unlabel A)"
  "setops\<^sub>s\<^sub>s\<^sub>t (proj_unl m (proj n A)) \<subseteq> setops\<^sub>s\<^sub>s\<^sub>t (proj_unl n A)"
  "setops\<^sub>s\<^sub>s\<^sub>t (proj_unl m (proj n A)) \<subseteq> setops\<^sub>s\<^sub>s\<^sub>t (proj_unl m A)"
unfolding unlabel_def proj_def
proof (induction A)
  case (Cons a A)
  obtain l b where lb: "a = (l,b)" by moura
  { case 1 thus ?case using Cons.IH lb by (cases b) auto }
  { case 2 thus ?case using Cons.IH lb by (cases b) auto }
  { case 3 thus ?case using Cons.IH lb by (cases b) auto }
qed simp_all

lemma setops\<^sub>s\<^sub>s\<^sub>t_prefix_subset:
  "setops\<^sub>s\<^sub>s\<^sub>t (unlabel A) \<subseteq> setops\<^sub>s\<^sub>s\<^sub>t (unlabel (A@B))"
  "setops\<^sub>s\<^sub>s\<^sub>t (proj_unl n A) \<subseteq> setops\<^sub>s\<^sub>s\<^sub>t (proj_unl n (A@B))"
unfolding unlabel_def proj_def
proof (induction A)
  case (Cons a A)
  obtain l b where lb: "a = (l,b)" by moura
  { case 1 thus ?case using Cons.IH lb by (cases b) auto }
  { case 2 thus ?case using Cons.IH lb by (cases b) auto }
qed simp_all

lemma setops\<^sub>s\<^sub>s\<^sub>t_suffix_subset:
  "setops\<^sub>s\<^sub>s\<^sub>t (unlabel B) \<subseteq> setops\<^sub>s\<^sub>s\<^sub>t (unlabel (A@B))"
  "setops\<^sub>s\<^sub>s\<^sub>t (proj_unl n B) \<subseteq> setops\<^sub>s\<^sub>s\<^sub>t (proj_unl n (A@B))"
unfolding unlabel_def proj_def
proof (induction A)
  case (Cons a A)
  obtain l b where lb: "a = (l,b)" by moura
  { case 1 thus ?case using Cons.IH lb by (cases b) auto }
  { case 2 thus ?case using Cons.IH lb by (cases b) auto }
qed simp_all

lemma setops\<^sub>l\<^sub>s\<^sub>s\<^sub>t_proj_subset:
  "setops\<^sub>l\<^sub>s\<^sub>s\<^sub>t (proj n A) \<subseteq> setops\<^sub>l\<^sub>s\<^sub>s\<^sub>t A"
  "setops\<^sub>l\<^sub>s\<^sub>s\<^sub>t (proj m (proj n A)) \<subseteq> setops\<^sub>l\<^sub>s\<^sub>s\<^sub>t (proj n A)"
unfolding proj_def by (induct A rule: setops\<^sub>l\<^sub>s\<^sub>s\<^sub>t.induct) auto

lemma setops\<^sub>l\<^sub>s\<^sub>s\<^sub>t_prefix_subset:
  "setops\<^sub>l\<^sub>s\<^sub>s\<^sub>t A \<subseteq> setops\<^sub>l\<^sub>s\<^sub>s\<^sub>t (A@B)"
  "setops\<^sub>l\<^sub>s\<^sub>s\<^sub>t (proj n A) \<subseteq> setops\<^sub>l\<^sub>s\<^sub>s\<^sub>t (proj n (A@B))"
unfolding proj_def by (induct A rule: setops\<^sub>l\<^sub>s\<^sub>s\<^sub>t.induct) auto

lemma setops\<^sub>l\<^sub>s\<^sub>s\<^sub>t_suffix_subset:
  "setops\<^sub>l\<^sub>s\<^sub>s\<^sub>t B \<subseteq> setops\<^sub>l\<^sub>s\<^sub>s\<^sub>t (A@B)"
  "setops\<^sub>l\<^sub>s\<^sub>s\<^sub>t (proj n B) \<subseteq> setops\<^sub>l\<^sub>s\<^sub>s\<^sub>t (proj n (A@B))"
unfolding proj_def by (induct A rule: setops\<^sub>l\<^sub>s\<^sub>s\<^sub>t.induct) auto

lemma par_comp\<^sub>l\<^sub>s\<^sub>s\<^sub>t_prefix:
  assumes "par_comp\<^sub>l\<^sub>s\<^sub>s\<^sub>t (A@B) Sec"
  shows "par_comp\<^sub>l\<^sub>s\<^sub>s\<^sub>t A Sec"
proof -
  let ?L = "\<lambda>n A. trms\<^sub>s\<^sub>s\<^sub>t (proj_unl n A) \<union> setops\<^sub>s\<^sub>s\<^sub>t (proj_unl n A)"

  have *: "\<And>n. ?L n A \<subseteq> ?L n (A@B)"
    using trms\<^sub>s\<^sub>s\<^sub>t_prefix_subset setops\<^sub>s\<^sub>s\<^sub>t_prefix_subset by fast
  have "\<And>m n. m \<noteq> n \<Longrightarrow> GSMP_disjoint (?L m A) (?L n A) Sec"
  proof -
    fix m n::'lbl assume "m \<noteq> n"
    thus "GSMP_disjoint (?L m A) (?L n A) Sec"
      using GSMP_disjoint_subset[of "?L m (A@B)" "?L n (A@B)" Sec "?L m A" "?L n A"] assms *
      unfolding par_comp\<^sub>l\<^sub>s\<^sub>s\<^sub>t_def by simp
  qed
  moreover have "setops\<^sub>l\<^sub>s\<^sub>s\<^sub>t A \<subseteq> setops\<^sub>l\<^sub>s\<^sub>s\<^sub>t (A@B)" using setops\<^sub>l\<^sub>s\<^sub>s\<^sub>t_prefix_subset by auto
  moreover have "set A \<subseteq> set (A@B)" by simp
  ultimately show ?thesis using assms unfolding par_comp\<^sub>l\<^sub>s\<^sub>s\<^sub>t_def by blast
qed

lemma par_comp\<^sub>l\<^sub>s\<^sub>s\<^sub>t_suffix:
  assumes "par_comp\<^sub>l\<^sub>s\<^sub>s\<^sub>t (A@B) Sec"
  shows "par_comp\<^sub>l\<^sub>s\<^sub>s\<^sub>t B Sec"
proof -
  let ?L = "\<lambda>n B. trms\<^sub>s\<^sub>s\<^sub>t (proj_unl n B) \<union> setops\<^sub>s\<^sub>s\<^sub>t (proj_unl n B)"

  have *: "\<And>n. ?L n B \<subseteq> ?L n (A@B)"
    using trms\<^sub>s\<^sub>s\<^sub>t_suffix_subset setops\<^sub>s\<^sub>s\<^sub>t_suffix_subset by fast
  have "\<And>m n. m \<noteq> n \<Longrightarrow> GSMP_disjoint (?L m B) (?L n B) Sec"
  proof -
    fix m n::'lbl assume "m \<noteq> n"
    thus "GSMP_disjoint (?L m B) (?L n B) Sec"
      using GSMP_disjoint_subset[of "?L m (A@B)" "?L n (A@B)" Sec "?L m B" "?L n B"] assms *
      unfolding par_comp\<^sub>l\<^sub>s\<^sub>s\<^sub>t_def by simp
  qed
  moreover have "setops\<^sub>l\<^sub>s\<^sub>s\<^sub>t B \<subseteq> setops\<^sub>l\<^sub>s\<^sub>s\<^sub>t (A@B)" using setops\<^sub>l\<^sub>s\<^sub>s\<^sub>t_suffix_subset by auto
  moreover have "set B \<subseteq> set (A@B)" by simp
  ultimately show ?thesis using assms unfolding par_comp\<^sub>l\<^sub>s\<^sub>s\<^sub>t_def by blast
qed

lemma par_comp\<^sub>l\<^sub>s\<^sub>s\<^sub>t_proj:
  assumes "par_comp\<^sub>l\<^sub>s\<^sub>s\<^sub>t A Sec"
  shows "par_comp\<^sub>l\<^sub>s\<^sub>s\<^sub>t (proj n A) Sec"
proof -
  define B where B: "B = proj n A"

  define M::"'lbl \<Rightarrow> ('fun,'var,'lbl) labeled_stateful_strand \<Rightarrow> ('fun,'var) terms" where M:
    "M = (\<lambda>(n::'lbl) A. trms\<^sub>s\<^sub>s\<^sub>t (proj_unl n A) \<union> setops\<^sub>s\<^sub>s\<^sub>t (proj_unl n A))"

  have "\<forall>t. (\<star>, Send t) \<notin> set A" "ground Sec" "\<forall>s\<in>Sec. \<forall>s'\<in>subterms s. {} \<turnstile>\<^sub>c s' \<or> s' \<in> Sec"
       "\<forall>(i,p)\<in>setops\<^sub>l\<^sub>s\<^sub>s\<^sub>t A. \<forall>(i',p')\<in>setops\<^sub>l\<^sub>s\<^sub>s\<^sub>t A.
          (\<exists>\<delta>. Unifier \<delta> (to_pair p) (to_pair p')) \<longrightarrow> i = i'"
    using assms unfolding par_comp\<^sub>l\<^sub>s\<^sub>s\<^sub>t_def by auto
  hence 1:
      "\<forall>t. (\<star>, Send t) \<notin> set B" "ground Sec" "\<forall>s\<in>Sec. \<forall>s'\<in>subterms s. {} \<turnstile>\<^sub>c s' \<or> s' \<in> Sec"
      "\<forall>(i,p)\<in>setops\<^sub>l\<^sub>s\<^sub>s\<^sub>t B. \<forall>(i',p')\<in>setops\<^sub>l\<^sub>s\<^sub>s\<^sub>t B.
          (\<exists>\<delta>. Unifier \<delta> (to_pair p) (to_pair p')) \<longrightarrow> i = i'"
    using proj_set_subset[of n A] setops\<^sub>l\<^sub>s\<^sub>s\<^sub>t_proj_subset(1)[of n A] B by blast+
  
  have 2: "\<forall>m n. m \<noteq> n \<longrightarrow> GSMP_disjoint (M n A) (M m A) Sec"
    using assms M unfolding par_comp\<^sub>l\<^sub>s\<^sub>s\<^sub>t_def by auto

  { fix m m'::'lbl assume "m \<noteq> m'"
    hence *: "GSMP_disjoint (M m' A) (M m A) Sec" using 2 by metis
    have **: "M m' B \<subseteq> M m' A" "M m B \<subseteq> M m A"
      using trms\<^sub>s\<^sub>s\<^sub>t_proj_subset(3)[of _ n A] setops\<^sub>s\<^sub>s\<^sub>t_proj_subset(3)[of _ n A]
      unfolding M B by blast+
    have "GSMP_disjoint (M m' B) (M m B) Sec" by (rule GSMP_disjoint_subset[OF * **])
  } hence 3: "\<forall>m n. n \<noteq> m \<longrightarrow> GSMP_disjoint (M n B) (M m B) Sec" by metis

  show ?thesis using 1 3 unfolding M B par_comp\<^sub>l\<^sub>s\<^sub>s\<^sub>t_def by blast
qed

lemma wf_to_pair_ineqs_map':
  assumes "wf\<^sub>s\<^sub>t X (unlabel A)"
  shows "wf\<^sub>s\<^sub>t X (unlabel (map (\<lambda>d. (i,\<nexists>Y\<langle>to_pair (t, s) \<doteq> to_pair (snd d)\<rangle>)) (dbproj i D)@A))"
using assms by (induct D) auto

lemma wf_to_pair_eqs_ineqs_map':
  fixes A::"('fun,'var,'lbl) labeled_strand"
  assumes "wf\<^sub>s\<^sub>t X (unlabel A)"
          "Di \<in> set (subseqs (dbproj i D))"
          "\<forall>(i,t,t') \<in> set D. fv t \<union> fv t' \<subseteq> X"
  shows "wf\<^sub>s\<^sub>t X (unlabel (
            (map (\<lambda>d. (i,\<langle>to_pair (t,s) \<doteq> to_pair (snd d)\<rangle>)) Di)@
            (map (\<lambda>d. (i,\<nexists>[]\<langle>to_pair (t,s) \<doteq> to_pair (snd d)\<rangle>)) [d\<leftarrow>dbproj i D. d \<notin> set Di])@A))"
proof -
  let ?f = "[d\<leftarrow>dbproj i D. d \<notin> set Di]"
  define c1 where c1: "c1 = map (\<lambda>d. (i,\<langle>to_pair (t,s) \<doteq> to_pair (snd d)\<rangle>)) Di"
  define c2 where c2: "c2 = map (\<lambda>d. (i,\<nexists>[]\<langle>to_pair (t,s) \<doteq> to_pair (snd d)\<rangle>)) ?f"
  define c3 where c3: "c3 = map (\<lambda>d. \<langle>to_pair (t,s) \<doteq> to_pair d\<rangle>) (unlabel Di)"
  define c4 where c4: "c4 = map (\<lambda>d. \<nexists>[]\<langle>to_pair (t,s) \<doteq> to_pair d\<rangle>) (unlabel ?f)"
  have ci_eqs: "c3 = unlabel c1" "c4 = unlabel c2" unfolding c1 c2 c3 c4 unlabel_def by auto
  have 1: "wf\<^sub>s\<^sub>t X (unlabel (c2@A))"
    using wf_to_pair_ineqs_map[OF assms(1)] ci_eqs(2) unlabel_append[of c2 A] c4
    by metis
  have 2: "\<forall>(i,t,t') \<in> set Di. fv t \<union> fv t' \<subseteq> X" 
    using assms(2,3) contra_subsetD subseqs_set_subset(1) by fastforce
  { fix B::"('fun,'var) strand" assume "wf\<^sub>s\<^sub>t X B"
    hence "wf\<^sub>s\<^sub>t X (unlabel c1@B)" using 2 unfolding c1 unlabel_def by (induct Di) auto
  } thus ?thesis using 1 unfolding c1 c2 unlabel_def by simp
qed


section {* Lemmata: Properties of the translation functions *}
lemma tr_par_labeled_rcv_iff:
  "B \<in> set (tr\<^sub>p\<^sub>c A D) \<Longrightarrow> (i, receive\<langle>t\<rangle>) \<in> set B \<longleftrightarrow> (i, Receive t) \<in> set A"
by (induct A D arbitrary: B rule: tr\<^sub>p\<^sub>c.induct) auto

lemma tr_par_declassified_eq:
  "B \<in> set (tr\<^sub>p\<^sub>c A D) \<Longrightarrow> declassified\<^sub>l\<^sub>s\<^sub>t B I = declassified\<^sub>l\<^sub>s\<^sub>s\<^sub>t A I"
using tr_par_labeled_rcv_iff unfolding declassified\<^sub>l\<^sub>s\<^sub>t_def declassified\<^sub>l\<^sub>s\<^sub>s\<^sub>t_def by simp

lemma tr_par_ik_eq:
  assumes "B \<in> set (tr\<^sub>p\<^sub>c A D)"
  shows "ik\<^sub>s\<^sub>t (unlabel B) = ik\<^sub>s\<^sub>s\<^sub>t (unlabel A)"
proof -
  have "{t. \<exists>i. (i, receive\<langle>t\<rangle>) \<in> set B} = {t. \<exists>i. (i, Receive t) \<in> set A}"
    using tr_par_labeled_rcv_iff[OF assms] by simp
  moreover have
      "\<And>C. {t. \<exists>i. (i, receive\<langle>t\<rangle>) \<in> set C} = {t. receive\<langle>t\<rangle> \<in> set (unlabel C)}"
      "\<And>C. {t. \<exists>i. (i, Receive t) \<in> set C} = {t. Receive t \<in> set (unlabel C)}"
    unfolding unlabel_def by force+
  ultimately show ?thesis by (metis ik\<^sub>s\<^sub>s\<^sub>t_def ik\<^sub>s\<^sub>t_is_rcv_set)
qed

lemma tr_par_deduct_iff:
  assumes "B \<in> set (tr\<^sub>p\<^sub>c A D)"
  shows "ik\<^sub>s\<^sub>t (unlabel B) \<cdot>\<^sub>s\<^sub>e\<^sub>t I \<turnstile> t \<longleftrightarrow> ik\<^sub>s\<^sub>s\<^sub>t (unlabel A) \<cdot>\<^sub>s\<^sub>e\<^sub>t I \<turnstile> t"
using tr_par_ik_eq[OF assms] by metis


lemma tr_par_vars_subset:
  assumes "A' \<in> set (tr\<^sub>p\<^sub>c A D)"
  shows "fv\<^sub>l\<^sub>s\<^sub>t A' \<subseteq> fv\<^sub>s\<^sub>s\<^sub>t (unlabel A) \<union> (\<Union>(l,t,t') \<in> set D. fv t \<union> fv t')" (is ?P)
  and "bvars\<^sub>l\<^sub>s\<^sub>t A' \<subseteq> bvars\<^sub>s\<^sub>s\<^sub>t (unlabel A)" (is ?Q)
proof -
  show ?P using assms
  proof (induction "unlabel A" arbitrary: A A' D rule: strand_sem_stateful_induct)
    case (ConsIn A' D t s AA A A')
    then obtain i B where iB: "A = (i,InSet t s)#B" "AA = unlabel B" unfolding unlabel_def by moura
    then obtain A'' d where *:
        "d \<in> set (dbproj i D)" "A' = (i,\<langle>to_pair (t,s) \<doteq> to_pair (snd d)\<rangle>)#A''" "A'' \<in> set (tr\<^sub>p\<^sub>c B D)"
      using ConsIn.prems(1) by moura
    hence "fv\<^sub>l\<^sub>s\<^sub>t A'' \<subseteq> fv\<^sub>s\<^sub>s\<^sub>t (unlabel B) \<union> (\<Union>(l,t,t')\<in>set D. fv t \<union> fv t')"
      by (metis ConsIn.hyps(1)[OF iB(2)])
    thus ?case using * iB by auto
  next
    case (ConsDel A' D t s AA A A')
    then obtain i B where iB: "A = (i,Delete t s)#B" "AA = unlabel B"
      unfolding unlabel_def by moura

    define Dfv where
      "Dfv = (\<lambda>D::('fun,'var,'lbl) labeleddbstatelist. (\<Union>(l,t,t')\<in>set D. fv t \<union> fv t'))"
    define fltD1 where "fltD1 = (\<lambda>Di. filter (\<lambda>d. d \<notin> set Di) D)"
    define fltD2 where "fltD2 = (\<lambda>Di. filter (\<lambda>d. d \<notin> set Di) (dbproj i D))"
    define constr where "constr =
      (\<lambda>Di. (map (\<lambda>d. (i, \<langle>to_pair (t,s) \<doteq> to_pair (snd d)\<rangle>)) Di)@
            (map (\<lambda>d. (i, \<nexists>[]\<langle>to_pair (t,s) \<doteq> to_pair (snd d)\<rangle>)) (fltD2 Di)))"

    from iB obtain A'' Di where *:
        "Di \<in> set (subseqs (dbproj i D))" "A' = (constr Di)@A''" "A'' \<in> set (tr\<^sub>p\<^sub>c B (fltD1 Di))"
      using ConsDel.prems(1) unfolding constr_def fltD1_def fltD2_def by moura
    hence "fv\<^sub>l\<^sub>s\<^sub>t A'' \<subseteq> fv\<^sub>s\<^sub>s\<^sub>t AA \<union> Dfv (filter (\<lambda>d. d \<notin> set Di) D)"
      unfolding constr_def fltD1_def Dfv_def by (metis ConsDel.hyps(1) iB(2))
    hence 1: "fv\<^sub>l\<^sub>s\<^sub>t A'' \<subseteq> fv\<^sub>s\<^sub>s\<^sub>t AA \<union> Dfv D" unfolding Dfv_def by auto
    
    have 2: "Dfv Di \<union> Dfv (fltD1 Di) \<subseteq> Dfv D"
      using subseqs_set_subset(1)[OF *(1)] unfolding fltD1_def Dfv_def by auto

    have 5: "fv\<^sub>l\<^sub>s\<^sub>t A' = fv\<^sub>l\<^sub>s\<^sub>t (constr Di) \<union> fv\<^sub>l\<^sub>s\<^sub>t A''" using * unfolding unlabel_def by force

    have "fv\<^sub>l\<^sub>s\<^sub>t (constr Di) \<subseteq> fv t \<union> fv s \<union> Dfv Di \<union> Dfv (fltD1 Di)"
      unfolding unlabel_def constr_def fltD1_def fltD2_def Dfv_def by auto
    hence 3: "fv\<^sub>l\<^sub>s\<^sub>t (constr Di) \<subseteq> fv t \<union> fv s \<union> Dfv D" using 2 by blast

    have 4: "fv\<^sub>s\<^sub>s\<^sub>t (unlabel A) = fv t \<union> fv s \<union> fv\<^sub>s\<^sub>s\<^sub>t AA" using iB by auto
    
    have "fv\<^sub>s\<^sub>t (unlabel A') \<subseteq> fv\<^sub>s\<^sub>s\<^sub>t (unlabel A) \<union> Dfv D" using 1 3 4 5 by blast
    thus ?case by (metis Dfv_def)
  qed (fastforce simp add: unlabel_def)+

  show ?Q using assms
    apply (induct "unlabel A" arbitrary: A A' D rule: strand_sem_stateful_induct)
    by (fastforce simp add: unlabel_def)+
qed

lemma tr_par_vars_disj:
  assumes "A' \<in> set (tr\<^sub>p\<^sub>c A D)" "\<forall>(l,t,t') \<in> set D. (fv t \<union> fv t') \<inter> bvars\<^sub>s\<^sub>s\<^sub>t (unlabel A) = {}"
  and "fv\<^sub>s\<^sub>s\<^sub>t (unlabel A) \<inter> bvars\<^sub>s\<^sub>s\<^sub>t (unlabel A) = {}"
  shows "fv\<^sub>l\<^sub>s\<^sub>t A' \<inter> bvars\<^sub>l\<^sub>s\<^sub>t A' = {}"
using assms tr_par_vars_subset by fast

lemma tr_par_trms_subset:
  assumes "A' \<in> set (tr\<^sub>p\<^sub>c A D)"
  shows "trms\<^sub>l\<^sub>s\<^sub>t A' \<subseteq> trms\<^sub>s\<^sub>s\<^sub>t (unlabel A) \<union> setops\<^sub>s\<^sub>s\<^sub>t (unlabel A) \<union> to_pair ` snd ` set D"
using assms
proof (induction A D arbitrary: A' rule: tr\<^sub>p\<^sub>c.induct)
  case 1 thus ?case by simp
next
  case (2 i t A D)
  then obtain A'' where A'': "A' = (i,send\<langle>t\<rangle>)#A''" "A'' \<in> set (tr\<^sub>p\<^sub>c A D)" by moura
  hence "trms\<^sub>l\<^sub>s\<^sub>t A'' \<subseteq> trms\<^sub>s\<^sub>s\<^sub>t (unlabel A) \<union> setops\<^sub>s\<^sub>s\<^sub>t (unlabel A) \<union> to_pair ` snd ` set D"
    by (metis "2.IH")
  thus ?case using A'' by auto
next
  case (3 i t A D)
  then obtain A'' where A'': "A' = (i,receive\<langle>t\<rangle>)#A''" "A'' \<in> set (tr\<^sub>p\<^sub>c A D)" by moura
  hence "trms\<^sub>l\<^sub>s\<^sub>t A'' \<subseteq> trms\<^sub>s\<^sub>s\<^sub>t (unlabel A) \<union> setops\<^sub>s\<^sub>s\<^sub>t (unlabel A) \<union> to_pair ` snd ` set D"
    by (metis "3.IH")
  thus ?case using A'' by auto
next
  case (4 i t t' A D)
  then obtain A'' where A'': "A' = (i,\<langle>t \<doteq> t'\<rangle>)#A''" "A'' \<in> set (tr\<^sub>p\<^sub>c A D)" by moura
  hence "trms\<^sub>l\<^sub>s\<^sub>t A'' \<subseteq> trms\<^sub>s\<^sub>s\<^sub>t (unlabel A) \<union> setops\<^sub>s\<^sub>s\<^sub>t (unlabel A) \<union> to_pair ` snd ` set D"
    by (metis "4.IH")
  thus ?case using A'' by auto
next
  case (5 i X t t' A D)
  then obtain A'' where A'': "A' = (i,\<nexists>X\<langle>t \<doteq> t'\<rangle>)#A''" "A'' \<in> set (tr\<^sub>p\<^sub>c A D)" by moura
  hence "trms\<^sub>l\<^sub>s\<^sub>t A'' \<subseteq> trms\<^sub>s\<^sub>s\<^sub>t (unlabel A) \<union> setops\<^sub>s\<^sub>s\<^sub>t (unlabel A) \<union> to_pair ` snd ` set D"
    by (metis "5.IH")
  thus ?case using A'' by auto
next
  case (6 i t s A D)
  hence "A' \<in> set (tr\<^sub>p\<^sub>c A (List.insert (i,t,s) D))" by simp
  hence "trms\<^sub>l\<^sub>s\<^sub>t A' \<subseteq> trms\<^sub>s\<^sub>s\<^sub>t (unlabel A) \<union> setops\<^sub>s\<^sub>s\<^sub>t (unlabel A) \<union>
                      to_pair ` snd ` set (List.insert (i,t,s) D)"
    by (metis "6.IH")
  thus ?case by auto
next
  case (7 i t s A D)
  from 7 obtain Di A'' B C where A'':
      "Di \<in> set (subseqs (dbproj i D))" "A'' \<in> set (tr\<^sub>p\<^sub>c A [d\<leftarrow>D. d \<notin> set Di])" "A' = (B@C)@A''"
      "B = map (\<lambda>d. (i,\<langle>to_pair (t,s) \<doteq> to_pair (snd d)\<rangle>)) Di"
      "C = map (\<lambda>d. (i,\<nexists>[]\<langle>to_pair (t,s) \<doteq> to_pair (snd d)\<rangle>)) [d\<leftarrow>dbproj i D. d \<notin> set Di]"
    by moura
  hence "trms\<^sub>l\<^sub>s\<^sub>t A'' \<subseteq> trms\<^sub>s\<^sub>s\<^sub>t (unlabel A) \<union> setops\<^sub>s\<^sub>s\<^sub>t (unlabel A) \<union>
                       to_pair ` snd ` set [d\<leftarrow>D. d \<notin> set Di]"
    by (metis "7.IH")
  moreover have "set [d\<leftarrow>D. d \<notin> set Di] \<subseteq> set D" using set_filter by auto
  ultimately have "trms\<^sub>l\<^sub>s\<^sub>t A'' \<subseteq> trms\<^sub>s\<^sub>s\<^sub>t (unlabel A) \<union> setops\<^sub>s\<^sub>s\<^sub>t (unlabel A) \<union> to_pair ` snd ` set D"
    by blast
  hence "trms\<^sub>l\<^sub>s\<^sub>t A'' \<subseteq> trms\<^sub>s\<^sub>s\<^sub>t (unlabel ((i,Delete t s)#A)) \<union>
                        setops\<^sub>s\<^sub>s\<^sub>t (unlabel ((i,Delete t s)#A)) \<union>
                        to_pair ` snd ` set D"
    using setops\<^sub>s\<^sub>s\<^sub>t_cons_subset trms\<^sub>s\<^sub>s\<^sub>t_cons
    by auto
  moreover have "set Di \<subseteq> set D" "set [d\<leftarrow>dbproj i D . d \<notin> set Di] \<subseteq> set D"
    using subseqs_set_subset[OF A''(1)] by auto
  hence "trms\<^sub>s\<^sub>t (unlabel B) \<subseteq> insert (to_pair (t, s)) (to_pair ` snd ` set D)"
        "trms\<^sub>s\<^sub>t (unlabel C) \<subseteq> insert (to_pair (t, s)) (to_pair ` snd ` set D)"
    using A''(4,5) unfolding unlabel_def by auto
  hence "trms\<^sub>s\<^sub>t (unlabel (B@C)) \<subseteq> insert (to_pair (t,s)) (to_pair ` snd ` set D)"
    using unlabel_append[of B C] by auto
  moreover have "to_pair (t,s) \<in> setops\<^sub>s\<^sub>s\<^sub>t (Delete t s#unlabel A)" by simp
  ultimately show ?case
    using A''(3) trms\<^sub>s\<^sub>t_append[of "unlabel (B@C)" "unlabel A'"] unlabel_append[of "B@C" A'']
    by auto
next
  case (8 i t s A D)
  from 8 obtain d A'' where A'':
      "d \<in> set (dbproj i D)" "A'' \<in> set (tr\<^sub>p\<^sub>c A D)" "A' = (i,\<langle>to_pair (t,s) \<doteq> to_pair (snd d)\<rangle>)#A''"
    by moura
  hence "trms\<^sub>l\<^sub>s\<^sub>t A'' \<subseteq> trms\<^sub>s\<^sub>s\<^sub>t (unlabel A) \<union> setops\<^sub>s\<^sub>s\<^sub>t (unlabel A) \<union>
                       to_pair ` snd ` set D"
    by (metis "8.IH")
  moreover have "trms\<^sub>s\<^sub>t (unlabel A') = {to_pair (t,s), to_pair (snd d)} \<union> trms\<^sub>s\<^sub>t (unlabel A'')"
    using A''(1,3) by auto
  ultimately show ?case using A''(1) by auto
next
  case (9 i X t s A D)
  define B where "B = (map (\<lambda>d. (i,\<nexists>X\<langle>to_pair (t,s) \<doteq> to_pair (snd d)\<rangle>)) (dbproj i D))"
  from 9 obtain A'' where A'':
      "A'' \<in> set (tr\<^sub>p\<^sub>c A D)" "A' = B@A''"
    unfolding B_def by moura
  hence "trms\<^sub>s\<^sub>t (unlabel A'') \<subseteq> trms\<^sub>s\<^sub>s\<^sub>t (unlabel A) \<union> setops\<^sub>s\<^sub>s\<^sub>t (unlabel A) \<union> to_pair ` snd ` set D"
    by (metis "9.IH")
  moreover have "trms\<^sub>s\<^sub>t (unlabel B) \<subseteq> {to_pair (t,s)} \<union> to_pair ` snd ` set D"
    unfolding unlabel_def B_def by auto
  ultimately have "trms\<^sub>s\<^sub>t (unlabel A') \<subseteq> {to_pair (t,s)} \<union> trms\<^sub>s\<^sub>s\<^sub>t (unlabel A) \<union>
                                         setops\<^sub>s\<^sub>s\<^sub>t (unlabel A) \<union> to_pair ` snd ` set D"
    using A'' unlabel_append[of B A''] by auto
  thus ?case by auto
qed

lemma tr_par_wf_trms:
  assumes "A' \<in> set (tr\<^sub>p\<^sub>c A [])" "wf\<^sub>t\<^sub>r\<^sub>m\<^sub>s (trms\<^sub>s\<^sub>s\<^sub>t (unlabel A))"
  shows "wf\<^sub>t\<^sub>r\<^sub>m\<^sub>s (trms\<^sub>l\<^sub>s\<^sub>t A')"
using tr_par_trms_subset[OF assms(1)] setops_wf_trms(2)[OF assms(2)]
by auto

lemma tr_par_wf':
  assumes "\<forall>(i,t,t') \<in> set D. (fv t \<union> fv t') \<inter> bvars\<^sub>s\<^sub>s\<^sub>t (unlabel A) = {}"
  and "\<forall>(i,t,t') \<in> set D. fv t \<union> fv t' \<subseteq> X"
  and "wf'\<^sub>s\<^sub>s\<^sub>t X (unlabel A)" "fv\<^sub>s\<^sub>s\<^sub>t (unlabel A) \<inter> bvars\<^sub>s\<^sub>s\<^sub>t (unlabel A) = {}"
  and "A' \<in> set (tr\<^sub>p\<^sub>c A D)"
  shows "wf\<^sub>l\<^sub>s\<^sub>t X A'"
proof -
  define P where
    "P = (\<lambda>(D::('fun,'var,'lbl) labeleddbstatelist) (A::('fun,'var,'lbl) labeled_stateful_strand).
          (\<forall>(i,t,t') \<in> set D. (fv t \<union> fv t') \<inter> bvars\<^sub>s\<^sub>s\<^sub>t (unlabel A) = {}) \<and>
          fv\<^sub>s\<^sub>s\<^sub>t (unlabel A) \<inter> bvars\<^sub>s\<^sub>s\<^sub>t (unlabel A) = {})"

  have "P D A" using assms(1,4) by (simp add: P_def)
  with assms(5,3,2) show ?thesis
  proof (induction A arbitrary: X A' D)
    case Nil thus ?case by simp
  next
    case (Cons a A)
    obtain i s where i: "a = (i,s)" by (metis surj_pair)
    note prems = Cons.prems
    note IH = Cons.IH
    show ?case
    proof (cases s)
      case (Receive t)
      note si = Receive i
      then obtain A'' where A'': "A' = (i,receive\<langle>t\<rangle>)#A''" "A'' \<in> set (tr\<^sub>p\<^sub>c A D)" "fv t \<subseteq> X"
        using prems unlabel_Cons[of i s A] by moura
      have *: "wf'\<^sub>s\<^sub>s\<^sub>t X (unlabel A)"
              "\<forall>(i,s,s') \<in> set D. fv s \<union> fv s' \<subseteq> X"
              "P D A"
        using prems si apply (force, force)
        using prems(4) si unfolding P_def by force
      show ?thesis using IH[OF A''(2) *] A''(1,3) by simp
    next
      case (Send t)
      note si = Send i
      then obtain A'' where A'': "A' = (i,send\<langle>t\<rangle>)#A''" "A'' \<in> set (tr\<^sub>p\<^sub>c A D)"
        using prems by moura
      have *: "wf'\<^sub>s\<^sub>s\<^sub>t (X \<union> fv t) (unlabel A)"
              "\<forall>(i,s,s') \<in> set D. fv s \<union> fv s' \<subseteq> X \<union> fv t"
              "P D A"
        using prems si apply (force, force)
        using prems(4) si unfolding P_def by force
      show ?thesis using IH[OF A''(2) *] A''(1) by simp
    next
      case (Equality t t')
      note si = Equality i
      then obtain A'' where A'': "A' = (i,\<langle>t \<doteq> t'\<rangle>)#A''" "A'' \<in> set (tr\<^sub>p\<^sub>c A D)" "fv t' \<subseteq> X"
        using prems unlabel_Cons[of i s] by moura
      have *: "wf'\<^sub>s\<^sub>s\<^sub>t (X \<union> fv t) (unlabel A)"
              "\<forall>(i,s,s') \<in> set D. fv s \<union> fv s' \<subseteq> X \<union> fv t"
              "P D A"
        using prems si apply (force, force)
        using prems(4) si unfolding P_def by force
      show ?thesis using IH[OF A''(2) *] A''(1,3) by simp 
    next
      case (Inequality Y t t')
      note si = Inequality i
      then obtain A'' where A'': "A' = (i,\<nexists>Y\<langle>t \<doteq> t'\<rangle>)#A''" "A'' \<in> set (tr\<^sub>p\<^sub>c A D)" using prems by moura
      have *: "wf'\<^sub>s\<^sub>s\<^sub>t X (unlabel A)"
              "\<forall>(i,s,s') \<in> set D. fv s \<union> fv s' \<subseteq> X"
        using prems si by auto
      have "bvars\<^sub>s\<^sub>s\<^sub>t (unlabel A) \<subseteq> bvars\<^sub>s\<^sub>s\<^sub>t (unlabel ((i,Inequality Y t t')#A))"
           "fv\<^sub>s\<^sub>s\<^sub>t (unlabel A) \<subseteq> fv\<^sub>s\<^sub>s\<^sub>t (unlabel ((i,Inequality Y t t')#A))"
        by auto
      hence **: "P D A" using prems si unfolding P_def by blast 
      show ?thesis using IH[OF A''(2) * **] A''(1) by simp
    next
      case (Insert t t')
      note si = Insert i
      hence A': "A' \<in> set (tr\<^sub>p\<^sub>c A (List.insert (i,t,t') D))" "fv t \<subseteq> X" "fv t' \<subseteq> X"
        using prems by auto
      have *: "wf'\<^sub>s\<^sub>s\<^sub>t X (unlabel A)" "\<forall>(i,s,s') \<in> set (List.insert (i,t,t') D). fv s \<union> fv s' \<subseteq> X"
        using prems si by auto
      have **: "P (List.insert (i,t,t') D) A" using prems(4) si unfolding P_def by force
      show ?thesis using IH[OF A'(1) * **] A'(2,3) by simp
    next
      case (Delete t t')
      note si = Delete i
      define constr where "constr = (\<lambda>Di.
        (map (\<lambda>d. (i,\<langle>to_pair (t,t') \<doteq> to_pair (snd d)\<rangle>)) Di)@
        (map (\<lambda>d. (i,\<nexists>[]\<langle>to_pair (t,t') \<doteq> to_pair (snd d)\<rangle>)) [d\<leftarrow>dbproj i D. d \<notin> set Di]))"
      from prems si obtain Di A'' where A'':
          "A' = constr Di@A''" "A'' \<in> set (tr\<^sub>p\<^sub>c A [d\<leftarrow>D. d \<notin> set Di])"
          "Di \<in> set (subseqs (dbproj i D))"
        unfolding constr_def by auto
      have *: "wf'\<^sub>s\<^sub>s\<^sub>t X (unlabel A)"
              "\<forall>(i,t',s') \<in> set (filter (\<lambda>d. d \<notin> set Di) D). fv t' \<union> fv s' \<subseteq> X"
        using prems si  by auto
      have **: "P [d\<leftarrow>D. d \<notin> set Di] A" using prems si unfolding P_def by force
      have ***: "\<forall>(i,t, t') \<in> set D. fv t \<union> fv t' \<subseteq> X" using prems si by auto
      show ?thesis
        using IH[OF A''(2) * **] A''(1) wf_to_pair_eqs_ineqs_map'[OF _ A''(3) ***]
        unfolding constr_def by simp
    next
      case (InSet t t')
      note si = InSet i
      then obtain d A'' where A'':
          "A' = (i,\<langle>to_pair (t,t') \<doteq> to_pair (snd d)\<rangle>)#A''"
          "A'' \<in> set (tr\<^sub>p\<^sub>c A D)"
          "d \<in> set D"
        using prems by moura
      have *:
          "wf'\<^sub>s\<^sub>s\<^sub>t (X \<union> fv t \<union> fv t') (unlabel A)"
          "\<forall>(i,u,u')\<in>set D. fv u \<union> fv u' \<subseteq> X \<union> fv t \<union> fv t'"
          "P D A"
        using prems si apply (force, force)
        using prems(4) si unfolding P_def by force
      have **: "fv (to_pair (snd d)) \<subseteq> X" using A''(3) prems(3) by fastforce
      have ***: "fv (to_pair (t,t')) = fv t \<union> fv t'" by auto
      show ?thesis using IH[OF A''(2) *] A''(1) ** *** by (simp add: Un_assoc)
    next
      case (NotInSet Y t t')
      note si = NotInSet i
      then obtain A'' where A'':
          "A' = (map (\<lambda>d. (i,\<nexists>Y\<langle>to_pair (t,t') \<doteq> to_pair (snd d)\<rangle>)) (dbproj i D))@A''"
          "A'' \<in> set (tr\<^sub>p\<^sub>c A D)"
        using prems by moura
      have *: "wf'\<^sub>s\<^sub>s\<^sub>t X (unlabel A)" "\<forall>(i,u,u') \<in> set D. fv u \<union> fv u' \<subseteq> X" using prems si by auto
      have "bvars\<^sub>s\<^sub>s\<^sub>t (unlabel A) \<subseteq> bvars\<^sub>s\<^sub>s\<^sub>t (unlabel ((i,NotInSet Y t t')#A))"
           "fv\<^sub>s\<^sub>s\<^sub>t (unlabel A) \<subseteq> fv\<^sub>s\<^sub>s\<^sub>t (unlabel ((i,NotInSet Y t t')#A))"
        by auto
      hence **:  "P D A" using prems si unfolding P_def by blast
      show ?thesis using IH[OF A''(2) * **] A''(1) wf_to_pair_ineqs_map' by simp
    qed
  qed
qed

lemma tr_par_wf:
  assumes "A' \<in> set (tr\<^sub>p\<^sub>c A [])"
    and "wf\<^sub>s\<^sub>s\<^sub>t (unlabel A)"
    and "wf\<^sub>t\<^sub>r\<^sub>m\<^sub>s (trms\<^sub>s\<^sub>s\<^sub>t (unlabel A))" 
  shows "wf\<^sub>l\<^sub>s\<^sub>t {} A'"
    and "wf\<^sub>t\<^sub>r\<^sub>m\<^sub>s (trms\<^sub>l\<^sub>s\<^sub>t A')"
    and "fv\<^sub>l\<^sub>s\<^sub>t A' \<inter> bvars\<^sub>l\<^sub>s\<^sub>t A' = {}"
using tr_par_wf'[OF _ _ _ _ assms(1)]
      tr_par_wf_trms[OF assms(1,3)]
      tr_par_vars_disj[OF assms(1)]
      assms(2)
by fastforce+

lemma tr_par_tfr\<^sub>s\<^sub>s\<^sub>t\<^sub>p:
  assumes "A' \<in> set (tr\<^sub>p\<^sub>c A D)" "list_all tfr\<^sub>s\<^sub>s\<^sub>t\<^sub>p (unlabel A)"
  and "fv\<^sub>s\<^sub>s\<^sub>t (unlabel A) \<inter> bvars\<^sub>s\<^sub>s\<^sub>t (unlabel A) = {}" (is "?P0 A D")
  and "\<forall>(i,t,s) \<in> set D. (fv t \<union> fv s) \<inter> bvars\<^sub>s\<^sub>s\<^sub>t (unlabel A) = {}" (is "?P1 A D")
  and "\<forall>t \<in> setops\<^sub>s\<^sub>s\<^sub>t (unlabel A) \<union> to_pair ` snd ` set D.
       \<forall>t' \<in> setops\<^sub>s\<^sub>s\<^sub>t (unlabel A) \<union> to_pair ` snd ` set D.
          (\<exists>\<delta>. Unifier \<delta> t t') \<longrightarrow> \<Gamma> t = \<Gamma> t'" (is "?P3 A D")
  shows "list_all tfr\<^sub>s\<^sub>t\<^sub>p (unlabel A')"
proof -
  have sublmm: "list_all tfr\<^sub>s\<^sub>s\<^sub>t\<^sub>p (unlabel A)" "?P0 A D" "?P1 A D" "?P3 A D"
    when p: "list_all tfr\<^sub>s\<^sub>s\<^sub>t\<^sub>p (unlabel (a#A))" "?P0 (a#A) D" "?P1 (a#A) D" "?P3 (a#A) D"
    for a A D
  proof -
    show "list_all tfr\<^sub>s\<^sub>s\<^sub>t\<^sub>p (unlabel A)" using p(1) by (simp add: unlabel_def tfr\<^sub>s\<^sub>s\<^sub>t_def)
    show "?P0 A D" using p(2) fv\<^sub>s\<^sub>s\<^sub>t_cons_subset unfolding unlabel_def by fastforce
    show "?P1 A D" using p(3) bvars\<^sub>s\<^sub>s\<^sub>t_cons_subset unfolding unlabel_def by fastforce
    have "setops\<^sub>s\<^sub>s\<^sub>t (unlabel A) \<subseteq> setops\<^sub>s\<^sub>s\<^sub>t (unlabel (a#A))"
      using setops\<^sub>s\<^sub>s\<^sub>t_cons_subset unfolding unlabel_def by auto
    thus "?P3 A D" using p(4) by blast
  qed

  show ?thesis using assms
  proof (induction A D arbitrary: A' rule: tr\<^sub>p\<^sub>c.induct)
    case 1 thus ?case by simp
  next
    case (2 i t A D)
    note prems = "2.prems"
    note IH = "2.IH"
    from prems(1) obtain A'' where A'': "A' = (i,send\<langle>t\<rangle>)#A''" "A'' \<in> set (tr\<^sub>p\<^sub>c A D)" by moura
    have "list_all tfr\<^sub>s\<^sub>t\<^sub>p (unlabel A'')"
      using IH[OF A''(2)] prems(5) sublmm[OF prems(2,3,4,5)]
      by meson
    thus ?case using A''(1) by simp
  next
    case (3 i t A D)
    note prems = "3.prems"
    note IH = "3.IH"
    from prems(1) obtain A'' where A'': "A' = (i,receive\<langle>t\<rangle>)#A''" "A'' \<in> set (tr\<^sub>p\<^sub>c A D)" by moura
    have "list_all tfr\<^sub>s\<^sub>t\<^sub>p (unlabel A'')"
      using IH[OF A''(2)] prems(5) sublmm[OF prems(2,3,4,5)]
      by meson
    thus ?case using A''(1) by simp
  next
    case (4 i t t' A D)
    note prems = "4.prems"
    note IH = "4.IH"
    from prems(1) obtain A'' where A'': "A' = (i,\<langle>t \<doteq> t'\<rangle>)#A''" "A'' \<in> set (tr\<^sub>p\<^sub>c A D)" by moura
    have "list_all tfr\<^sub>s\<^sub>t\<^sub>p (unlabel A'')"
      using IH[OF A''(2)] prems(5) sublmm[OF prems(2,3,4,5)]
      by meson
    thus ?case using A''(1) prems(2) by simp
  next
    case (5 i X t t' A D)
    note prems = "5.prems"
    note IH = "5.IH"
    from prems(1) obtain A'' where A'': "A' = (i,\<nexists>X\<langle>t \<doteq> t'\<rangle>)#A''" "A'' \<in> set (tr\<^sub>p\<^sub>c A D)" by moura
    have "list_all tfr\<^sub>s\<^sub>t\<^sub>p (unlabel A'')"
      using IH[OF A''(2)] prems(5) sublmm[OF prems(2,3,4,5)]
      by meson
    thus ?case using A''(1) prems(2) by simp
  next
    case (6 i t s A D)
    note prems = "6.prems"
    note IH = "6.IH"
    from prems(1) have A': "A' \<in> set (tr\<^sub>p\<^sub>c A (List.insert (i,t,s) D))" by simp

    have 1: "list_all tfr\<^sub>s\<^sub>s\<^sub>t\<^sub>p (unlabel A)" using sublmm[OF prems(2,3,4,5)] by simp
  
    have "setops\<^sub>s\<^sub>s\<^sub>t (unlabel ((i,Insert t s)#A)) \<union> to_pair`snd`set D =
          setops\<^sub>s\<^sub>s\<^sub>t (unlabel A) \<union> to_pair`snd`set (List.insert (i,t,s) D)"
      by auto
    hence 3: "?P3 A (List.insert (i,t,s) D)" using prems(5) by metis
    moreover have "?P1 A (List.insert (i,t,s) D)"
      using prems(3,4) bvars\<^sub>s\<^sub>s\<^sub>t_cons_subset[of "unlabel A"] by auto
    ultimately have "list_all tfr\<^sub>s\<^sub>t\<^sub>p (unlabel A')"
      using IH[OF A' sublmm(1,2)[OF prems(2,3,4,5)] _ 3] by metis
    thus ?case using A'(1) by auto
  next
    case (7 i t s A D)
    note prems = "7.prems"
    note IH = "7.IH"

    define constr where constr: "constr \<equiv> (\<lambda>Di.
      (map (\<lambda>d. (i,\<langle>to_pair (t,s) \<doteq> to_pair (snd d)\<rangle>)) Di)@
      (map (\<lambda>d. (i,\<nexists>[]\<langle>to_pair (t,s) \<doteq> to_pair (snd d)\<rangle>)) (filter (\<lambda>d. d \<notin> set Di) (dbproj i D))))"

    from prems(1) obtain Di A'' where A'':
        "A' = constr Di@A''" "A'' \<in> set (tr\<^sub>p\<^sub>c A (filter (\<lambda>d. d \<notin> set Di) D))"
        "Di \<in> set (subseqs (dbproj i D))"
      unfolding constr by fastforce

    define Q1 where "Q1 \<equiv> (\<lambda>(t::('fun,'var) term) (t'::('fun,'var) term) X.
        \<forall>x \<in> (fv t \<union> fv t') - set X. \<exists>a. \<Gamma> (Var x) = TAtom a)"
    define Q2 where "Q2 \<equiv> (\<lambda>(t::('fun,'var) term) (t'::('fun,'var) term) X.
        \<forall>f T. Fun f T \<in> subterms t \<union> subterms t' \<longrightarrow> T = [] \<or> (\<exists>s \<in> set T. s \<notin> Var ` set X))"

    have "set (filter (\<lambda>d. d \<notin> set Di) D) \<subseteq> set D"
         "setops\<^sub>s\<^sub>s\<^sub>t (unlabel A) \<union> to_pair`snd`set Di
            \<subseteq> setops\<^sub>s\<^sub>s\<^sub>t (unlabel ((i,Delete t s)#A)) \<union> to_pair`snd`set D"
      using subseqs_set_subset[OF A''(3)] by force+
    hence *: "?P3 A (filter (\<lambda>d. d \<notin> set Di) D)" using prems(5) by blast
    have **: "?P1 A (filter (\<lambda>d. d \<notin> set Di) D)"
      using prems(4,5) by auto
    have 1: "list_all tfr\<^sub>s\<^sub>t\<^sub>p (unlabel A'')"
      using IH[OF A''(3,2) sublmm(1,2)[OF prems(2,3,4,5)] ** *]
      by metis

    have 2: "\<And>u u'. \<langle>u \<doteq> u'\<rangle> \<in> set (unlabel A') \<Longrightarrow> \<langle>u \<doteq> u'\<rangle> \<in> set (unlabel A'') \<or>
                      (\<exists>d \<in> set Di. u = to_pair (t,s) \<and> u' = to_pair (snd d))"
      using A''(1) unfolding constr unlabel_def by force
    have 3:
        "\<And>X u u'. \<nexists>X\<langle>u \<doteq> u'\<rangle> \<in> set (unlabel A') \<Longrightarrow> \<nexists>X\<langle>u \<doteq> u'\<rangle> \<in> set (unlabel A'') \<or>
            (\<exists>d \<in> set (filter (\<lambda>d. d \<notin> set Di) D). u = to_pair (t,s) \<and> u' = to_pair (snd d) \<and> Q2 u u' X)"
      using A''(1) unfolding Q2_def constr unlabel_def by force
    have 4: "\<forall>d\<in>set D. (\<exists>\<delta>. Unifier \<delta> (to_pair (t,s)) (to_pair (snd d)))
                       \<longrightarrow> \<Gamma> (to_pair (t,s)) = \<Gamma> (to_pair (snd d))"
      using prems(5) by simp

    { fix u u'
      assume a: "\<langle>u \<doteq> u'\<rangle> \<in> set (unlabel A')" "\<exists>\<delta>. Unifier \<delta> u u'"
      hence "\<langle>u \<doteq> u'\<rangle> \<in> set (unlabel A'') \<or> (\<exists>d \<in> set Di. u = to_pair (t,s) \<and> u' = to_pair (snd d))"
        using 2 by metis
      moreover {
        assume "\<langle>u \<doteq> u'\<rangle> \<in> set (unlabel A'')"
        hence "tfr\<^sub>s\<^sub>t\<^sub>p \<langle>u \<doteq> u'\<rangle>" using 1 Ball_set_list_all[of "unlabel A''" tfr\<^sub>s\<^sub>t\<^sub>p] by fast
      } moreover {
        fix d assume "d \<in> set Di" "u = to_pair (t,s)" "u' = to_pair (snd d)"
        hence "\<exists>\<delta>. Unifier \<delta> u u' \<Longrightarrow> \<Gamma> u = \<Gamma> u'" using 4 dbproj_subseq_subset A''(3) by fast
        hence "tfr\<^sub>s\<^sub>t\<^sub>p \<langle>u \<doteq> u'\<rangle>" using Ball_set_list_all[of "unlabel A''" tfr\<^sub>s\<^sub>t\<^sub>p]by simp
        hence "\<Gamma> u = \<Gamma> u'" using tfr\<^sub>s\<^sub>t\<^sub>p_list_all_alt_def[of "unlabel A''"]
          using a(2) unfolding unlabel_def by auto
      } ultimately have "\<Gamma> u = \<Gamma> u'"
          using tfr\<^sub>s\<^sub>t\<^sub>p_list_all_alt_def[of "unlabel A''"] a(2)
          unfolding unlabel_def by auto
    } moreover {
      fix u u' U
      assume "\<nexists>U\<langle>u \<doteq> u'\<rangle> \<in> set (unlabel A')"
      hence "\<nexists>U\<langle>u \<doteq> u'\<rangle> \<in> set (unlabel A'') \<or>
             (\<exists>d \<in> set (filter (\<lambda>d. d \<notin> set Di) D). u = to_pair (t,s) \<and> u' = to_pair (snd d) \<and> Q2 u u' U)"
        using 3 by metis
      hence "Q1 u u' U \<or> Q2 u u' U"
        using 1 4 subseqs_set_subset[OF A''(3)] tfr\<^sub>s\<^sub>t\<^sub>p_list_all_alt_def[of "unlabel A''"]
        unfolding Q1_def Q2_def
        by blast
    } ultimately show ?case
      using tfr\<^sub>s\<^sub>t\<^sub>p_list_all_alt_def[of "unlabel A'"] unfolding Q1_def Q2_def unlabel_def by blast
  next
    case (8 i t s A D)
    note prems = "8.prems"
    note IH = "8.IH"

    from prems(1) obtain d A'' where A'':
        "A' = (i,\<langle>to_pair (t,s) \<doteq> to_pair (snd d)\<rangle>)#A''" "A'' \<in> set (tr\<^sub>p\<^sub>c A D)"
        "d \<in> set (dbproj i D)"
      by moura

    have 1: "list_all tfr\<^sub>s\<^sub>t\<^sub>p (unlabel A'')"
      using IH[OF A''(2) sublmm(1,2,3)[OF prems(2,3,4,5)] sublmm(4)[OF prems(2,3,4,5)]] 
      by metis
    
    have 2: "(\<exists>\<delta>. Unifier \<delta> (to_pair (t,s)) (to_pair (snd d)))
                    \<Longrightarrow> \<Gamma> (to_pair (t,s)) = \<Gamma> (to_pair (snd d))"
      using prems(2,5) A''(3) unfolding tfr\<^sub>s\<^sub>s\<^sub>t_def by simp
    
    show ?case using A''(1) 1 2 by fastforce
  next
    case (9 i X t s A D)
    note prems = "9.prems"
    note IH = "9.IH"

    define constr where "constr = map (\<lambda>d. (i,\<nexists>X\<langle>to_pair (t,s) \<doteq> to_pair (snd d)\<rangle>)) (dbproj i D)"

    define Q2 where "Q2 \<equiv> (\<lambda>(t::('fun,'var) term) X.
        \<forall>f T. Fun f T \<in> subterms t \<longrightarrow> T = [] \<or> (\<exists>s \<in> set T. s \<notin> Var ` set X))"
  
    from prems(1) obtain A'' where A'': "A' = constr@A''" "A'' \<in> set (tr\<^sub>p\<^sub>c A D)"
      unfolding constr_def by moura

    have 0: "Q2 (to_pair (t,s)) X"
      using prems(2) unfolding Q2_def by auto

    have 1: "list_all tfr\<^sub>s\<^sub>t\<^sub>p (unlabel A'')"
      using IH[OF A''(2) sublmm(1,2,3)[OF prems(2,3,4,5)] sublmm(4)[OF prems(2,3,4,5)]]
      by metis

    { fix c assume "c \<in> set (unlabel constr)"
      hence "\<exists>d \<in> set (dbproj i D). c = \<nexists>X\<langle>to_pair (t,s) \<doteq> to_pair (snd d)\<rangle>"
        unfolding constr_def unlabel_def by force
    } moreover {
      fix d assume "d \<in> set (dbproj i D)"
      hence "fv (to_pair (snd d)) \<inter> set X = {}" using prems(4) by force
      hence "Q2 (to_pair (snd d)) X" by (metis fv_disj_Fun_subterm_param_cases Q2_def)
    } ultimately have 2: "list_all tfr\<^sub>s\<^sub>t\<^sub>p (unlabel constr)"
      using tfr\<^sub>s\<^sub>t\<^sub>p_list_all_alt_def[of "unlabel constr"] 0 unfolding Q2_def by blast
  
    show ?case using 1 2 A''(1) unfolding unlabel_def by simp
  qed
qed

lemma tr_par_tfr:
  assumes "A' \<in> set (tr\<^sub>p\<^sub>c A [])" and "tfr\<^sub>s\<^sub>s\<^sub>t (unlabel A)"
    and "fv\<^sub>s\<^sub>s\<^sub>t (unlabel A) \<inter> bvars\<^sub>s\<^sub>s\<^sub>t (unlabel A) = {}"
  shows "tfr\<^sub>s\<^sub>t (unlabel A')"
proof -
  have *: "trms\<^sub>l\<^sub>s\<^sub>t A' \<subseteq> trms\<^sub>s\<^sub>s\<^sub>t (unlabel A) \<union> setops\<^sub>s\<^sub>s\<^sub>t (unlabel A)"
    using tr_par_trms_subset[OF assms(1)] by simp
  hence "SMP (trms\<^sub>l\<^sub>s\<^sub>t A') \<subseteq> SMP (trms\<^sub>s\<^sub>s\<^sub>t (unlabel A) \<union> setops\<^sub>s\<^sub>s\<^sub>t (unlabel A))" using SMP_mono by simp
  moreover have "tfr\<^sub>s\<^sub>e\<^sub>t (trms\<^sub>s\<^sub>s\<^sub>t (unlabel A) \<union> setops\<^sub>s\<^sub>s\<^sub>t (unlabel A))"
    using assms(2) unfolding tfr\<^sub>s\<^sub>s\<^sub>t_def by fast
  ultimately have 1: "tfr\<^sub>s\<^sub>e\<^sub>t (trms\<^sub>l\<^sub>s\<^sub>t A')" by (metis tfr_subset[OF _ *])

  have **: "list_all tfr\<^sub>s\<^sub>s\<^sub>t\<^sub>p (unlabel A)" using assms(2) unfolding tfr\<^sub>s\<^sub>s\<^sub>t_def by fast
  have "setops\<^sub>s\<^sub>s\<^sub>t (unlabel A) \<subseteq> SMP (trms\<^sub>s\<^sub>s\<^sub>t (unlabel A) \<union> setops\<^sub>s\<^sub>s\<^sub>t (unlabel A)) - Var`\<V>"
    using setops\<^sub>s\<^sub>s\<^sub>t_are_pairs by auto
  hence ***:
      "\<forall>t \<in> setops\<^sub>s\<^sub>s\<^sub>t (unlabel A). \<forall>t' \<in> setops\<^sub>s\<^sub>s\<^sub>t (unlabel A). (\<exists>\<delta>. Unifier \<delta> t t') \<longrightarrow> \<Gamma> t = \<Gamma> t'"
    using assms(2) unfolding tfr\<^sub>s\<^sub>s\<^sub>t_def tfr\<^sub>s\<^sub>e\<^sub>t_def by blast
  have 2: "list_all tfr\<^sub>s\<^sub>t\<^sub>p (unlabel A')"
    using tr_par_tfr\<^sub>s\<^sub>s\<^sub>t\<^sub>p[OF assms(1) ** assms(3)] *** by fastforce

  show ?thesis by (metis 1 2 tfr\<^sub>s\<^sub>t_def)
qed

lemma tr_par_proj:
  assumes "B \<in> set (tr\<^sub>p\<^sub>c A D)"
  shows "proj n B \<in> set (tr\<^sub>p\<^sub>c (proj n A) (proj n D))"
using assms
proof (induction A D arbitrary: B rule: tr\<^sub>p\<^sub>c.induct)
  case (6 i t s S D)
  note prems = "6.prems"
  note IH = "6.IH"
  have IH': "proj n B \<in> set (tr\<^sub>p\<^sub>c (proj n S) (proj n (List.insert (i,t,s) D)))" using prems IH by auto
  show ?case
  proof (cases "(i = ln n) \<or> (i = \<star>)")
    case True thus ?thesis
      using IH' proj_list_insert(1,2)[of n "(t,s)" D] proj_list_Cons(1,2)[of n _ S]
      by auto
  next
    case False
    then obtain m where "i = ln m" "n \<noteq> m" by (cases i) simp_all
    thus ?thesis
      using IH' proj_list_insert(3)[of n _ "(t,s)" D] proj_list_Cons(3)[of n _ "Insert t s" S]
      by auto
  qed
next
  case (7 i t s S D)
  note prems = "7.prems"
  note IH = "7.IH"
  define constr where "constr = (\<lambda>Di D.
      (map (\<lambda>d. (i,\<langle>to_pair (t,s) \<doteq> to_pair (snd d)\<rangle>)) Di)@
      (map (\<lambda>d. (i,\<nexists>[]\<langle>to_pair (t,s) \<doteq> to_pair (snd d)\<rangle>)) [d\<leftarrow>dbproj i D. d \<notin> set Di]))"

  obtain Di B' where B':
      "B = constr Di D@B'"
      "Di \<in> set (subseqs (dbproj i D))"
      "B' \<in> set (tr\<^sub>p\<^sub>c S [d\<leftarrow>D. d \<notin> set Di])"
    using prems constr_def by fastforce
  hence "proj n B' \<in> set (tr\<^sub>p\<^sub>c (proj n S) (proj n [d\<leftarrow>D. d \<notin> set Di]))" using IH by simp
  hence IH': "proj n B' \<in> set (tr\<^sub>p\<^sub>c (proj n S) [d\<leftarrow>proj n D. d \<notin> set Di])" by (metis proj_filter)
  show ?case
  proof (cases "(i = ln n) \<or> (i = \<star>)")
    case True
    hence "proj n B = constr Di D@proj n B'" "Di \<in> set (subseqs (dbproj i (proj n D)))"
      using B'(1,2) proj_dbproj(1,2)[of n D] unfolding proj_def constr_def by auto
    moreover have "constr Di (proj n D) = constr Di D"
      using True proj_dbproj(1,2)[of n D] unfolding constr_def by presburger
    ultimately have "proj n B \<in> set (tr\<^sub>p\<^sub>c ((i, Delete t s)#proj n S) (proj n D))"
      using IH' unfolding constr_def by force
    thus ?thesis by (metis proj_list_Cons(1,2) True)
  next
    case False
    then obtain m where m: "i = ln m" "n \<noteq> m" by (cases i) simp_all
    hence *: "(ln n) \<noteq> i" by simp
    have "proj n B = proj n B'" using B'(1) False unfolding constr_def proj_def by auto
    moreover have "[d\<leftarrow>proj n D. d \<notin> set Di] = proj n D"
      using proj_subseq[OF _ m(2)[symmetric]] m(1) B'(2) by simp
    ultimately show ?thesis using m(1) IH' proj_list_Cons(3)[OF m(2), of _ S] by auto
  qed
next
  case (8 i t s S D)
  note prems = "8.prems"
  note IH = "8.IH"
  define constr where "constr = (
    \<lambda>d::'lbl strand_label \<times> ('fun,'var) term \<times> ('fun,'var) term.
      (i,\<langle>to_pair (t,s) \<doteq> to_pair (snd d)\<rangle>))"

  obtain d B' where B':
      "B = constr d#B'"
      "d \<in> set (dbproj i D)"
      "B' \<in> set (tr\<^sub>p\<^sub>c S D)"
    using prems constr_def by fastforce
  hence IH': "proj n B' \<in> set (tr\<^sub>p\<^sub>c (proj n S) (proj n D))" using IH by auto

  show ?case
  proof (cases "(i = ln n) \<or> (i = \<star>)")
    case True
    hence "proj n B = constr d#proj n B'" "d \<in> set (dbproj i (proj n D))"
      using B' proj_list_Cons(1,2)[of n _ B']
      unfolding constr_def
      by (force, metis proj_dbproj(1,2))
    hence "proj n B \<in> set (tr\<^sub>p\<^sub>c ((i, InSet t s)#proj n S) (proj n D))"
      using IH' unfolding constr_def by auto
    thus ?thesis using proj_list_Cons(1,2)[of n _ S] True by metis
  next
    case False
    then obtain m where m: "i = ln m" "n \<noteq> m" by (cases i) simp_all
    hence "proj n B = proj n B'" using B'(1) proj_list_Cons(3) unfolding constr_def by auto
    thus ?thesis
      using IH' m proj_list_Cons(3)[OF m(2), of "InSet t s" S]
      unfolding constr_def
      by auto
  qed
next
  case (9 i X t s S D)
  note prems = "9.prems"
  note IH = "9.IH"
  define constr where "constr = (\<lambda>D.
    map (\<lambda>d. (i,\<nexists>X\<langle>to_pair (t,s) \<doteq> to_pair (snd d)\<rangle>)) (dbproj i D))"
  
  obtain B' where B':
      "B = constr D@B'"
      "B' \<in> set (tr\<^sub>p\<^sub>c S D)"
    using prems constr_def by fastforce
  hence IH': "proj n B' \<in> set (tr\<^sub>p\<^sub>c (proj n S) (proj n D))" using IH by auto

  show ?case
  proof (cases "(i = ln n) \<or> (i = \<star>)")
    case True
    hence "proj n B = constr (proj n D)@proj n B'"
      using B'(1,2) proj_dbproj(1,2)[of n D] unfolding proj_def constr_def by auto
    hence "proj n B \<in> set (tr\<^sub>p\<^sub>c ((i, NotInSet X t s)#proj n S) (proj n D))"
      using IH' unfolding constr_def by auto
    thus ?thesis using proj_list_Cons(1,2)[of n _ S] True by metis
  next
    case False
    then obtain m where m: "i = ln m" "n \<noteq> m" by (cases i) simp_all
    hence "proj n B = proj n B'" using B'(1) unfolding constr_def proj_def by auto
    thus ?thesis
      using IH' m proj_list_Cons(3)[OF m(2), of "NotInSet X t s" S]
      unfolding constr_def
      by auto
  qed
qed (force simp add: proj_def)+

lemma tr_par_preserves_typing_cond:
  assumes "par_comp\<^sub>l\<^sub>s\<^sub>s\<^sub>t A Sec" "typing_cond\<^sub>s\<^sub>s\<^sub>t (unlabel A)" "A' \<in> set (tr\<^sub>p\<^sub>c A [])"
  shows "typing_cond (unlabel A')"
proof -
  have "wf'\<^sub>s\<^sub>s\<^sub>t {} (unlabel A)"
       "fv\<^sub>s\<^sub>s\<^sub>t (unlabel A) \<inter> bvars\<^sub>s\<^sub>s\<^sub>t (unlabel A) = {}"
       "wf\<^sub>t\<^sub>r\<^sub>m\<^sub>s (trms\<^sub>s\<^sub>s\<^sub>t (unlabel A))"
    using assms(2) unfolding typing_cond\<^sub>s\<^sub>s\<^sub>t_def by simp_all
  hence 1: "wf\<^sub>s\<^sub>t {} (unlabel A')"
           "fv\<^sub>s\<^sub>t (unlabel A') \<inter> bvars\<^sub>s\<^sub>t (unlabel A') = {}"
           "wf\<^sub>t\<^sub>r\<^sub>m\<^sub>s (trms\<^sub>s\<^sub>t (unlabel A'))"
           "Ana_invar_subst (ik\<^sub>s\<^sub>t (unlabel A') \<union> eqs_rhs\<^sub>s\<^sub>t (unlabel A'))"
    using tr_par_wf[OF assms(3)] Ana_invar_subst' by metis+

  have 2: "tfr\<^sub>s\<^sub>t (unlabel A')" by (metis tr_par_tfr assms(2,3) typing_cond\<^sub>s\<^sub>s\<^sub>t_def)

  show ?thesis by (metis 1 2 typing_cond_def)
qed

lemma tr_par_preserves_par_comp:
  assumes "par_comp\<^sub>l\<^sub>s\<^sub>s\<^sub>t A Sec" "A' \<in> set (tr\<^sub>p\<^sub>c A [])"
  shows "par_comp A' Sec"
proof -
  let ?M = "\<lambda>l. trms\<^sub>s\<^sub>s\<^sub>t (proj_unl l A) \<union> setops\<^sub>s\<^sub>s\<^sub>t (proj_unl l A)"
  let ?N = "\<lambda>l. trms_proj\<^sub>l\<^sub>s\<^sub>t l A'"

  have 0: "\<forall>l1 l2. l1 \<noteq> l2 \<longrightarrow> GSMP_disjoint (?M l1) (?M l2) Sec"
          "\<forall>t. (\<star>, stateful_strand_step.Send t) \<notin> set A"
    using assms(1) unfolding par_comp\<^sub>l\<^sub>s\<^sub>s\<^sub>t_def by simp_all

  { fix l1 l2::'lbl assume *: "l1 \<noteq> l2"
    hence "GSMP_disjoint (?M l1) (?M l2) Sec" using 0(1) by metis
    moreover have "\<And>n. to_pair ` snd ` set (proj n []) = {}" unfolding proj_def by simp
    hence "?N l1 \<subseteq> ?M l1" "?N l2 \<subseteq> ?M l2"
      using tr_par_trms_subset[OF tr_par_proj[OF assms(2)]] by (metis Un_empty_right)+
    ultimately have "GSMP_disjoint (?N l1) (?N l2) Sec"
      by (metis GSMP_disjoint_subset)
  } hence 1: "\<forall>l1 l2. l1 \<noteq> l2 \<longrightarrow> GSMP_disjoint (trms_proj\<^sub>l\<^sub>s\<^sub>t l1 A') (trms_proj\<^sub>l\<^sub>s\<^sub>t l2 A') Sec"
    using 0(1) by metis

  { fix D assume "A' \<in> set (tr\<^sub>p\<^sub>c A D)"
    hence "\<forall>t. (\<star>, send\<langle>t\<rangle>) \<notin> set A'"
      using 0(2) by (induct A D arbitrary: A' rule: tr\<^sub>p\<^sub>c.induct) auto
  } hence 2: "\<forall>t. (\<star>, send\<langle>t\<rangle>) \<notin> set A'" using assms(2) by metis

  have 3: "ground Sec" "\<forall>s \<in> Sec. \<forall>s' \<in> subterms s. {} \<turnstile>\<^sub>c s' \<or> s' \<in> Sec"
    using assms(1) unfolding par_comp\<^sub>l\<^sub>s\<^sub>s\<^sub>t_def by metis+

  show ?thesis using 1 2 3 unfolding par_comp_def by metis
qed

lemma tr_leaking_prefix_exists:
  assumes "A' \<in> set (tr\<^sub>p\<^sub>c A [])" "prefix B A'" "ik\<^sub>s\<^sub>t (proj_unl n B) \<cdot>\<^sub>s\<^sub>e\<^sub>t \<I> \<turnstile> t \<cdot> \<I>"
  shows "\<exists>C D. prefix C B \<and> prefix D A \<and> C \<in> set (tr\<^sub>p\<^sub>c D []) \<and> (ik\<^sub>s\<^sub>t (proj_unl n C) \<cdot>\<^sub>s\<^sub>e\<^sub>t \<I> \<turnstile> t \<cdot> \<I>)"
proof -
  let ?P = "\<lambda>B C C'. B = C@C' \<and> (\<forall>n t. (n, receive\<langle>t\<rangle>) \<notin> set C') \<and>
                     (C = [] \<or> (\<exists>n t. suffix [(n,receive\<langle>t\<rangle>)] C))"
  have "\<exists>C C'. ?P B C C'"
  proof (induction B)
    case (Cons b B)
    then obtain C C' n s where *: "?P B C C'" "b = (n,s)" by moura
    show ?case
    proof (cases "C = []")
      case True
      note T = True
      show ?thesis
      proof (cases "\<exists>t. s = receive\<langle>t\<rangle>")
        case True
        hence "?P (b#B) [b] C'" using * T  by auto
        thus ?thesis by metis
      next
        case False
        hence "?P (b#B) [] (b#C')" using * T by auto
        thus ?thesis by metis
      qed
    next
      case False
      hence "?P (b#B) (b#C) C'" using * unfolding suffix_def by auto
      thus ?thesis by metis
    qed
  qed simp
  then obtain C C' where C:
      "B = C@C'" "\<forall>n t. (n, receive\<langle>t\<rangle>) \<notin> set C'"
      "C = [] \<or> (\<exists>n t. suffix [(n,receive\<langle>t\<rangle>)] C)"
    by moura
  hence 1: "prefix C B" by simp
  hence 2: "prefix C A'" using assms(2) by simp

  have "\<And>m t. (m,receive\<langle>t\<rangle>) \<in> set B \<Longrightarrow> (m,receive\<langle>t\<rangle>) \<in> set C" using C by auto
  hence "\<And>t. receive\<langle>t\<rangle> \<in> set (proj_unl n B) \<Longrightarrow> receive\<langle>t\<rangle> \<in> set (proj_unl n C)"
    unfolding unlabel_def proj_def by force
  hence "ik\<^sub>s\<^sub>t (proj_unl n B) \<subseteq> ik\<^sub>s\<^sub>t (proj_unl n C)" using ik\<^sub>s\<^sub>t_is_rcv_set by auto
  hence 3: "ik\<^sub>s\<^sub>t (proj_unl n C) \<cdot>\<^sub>s\<^sub>e\<^sub>t \<I> \<turnstile> t \<cdot> \<I>" by (metis ideduct_mono[OF assms(3)] subst_all_mono)

  { fix D E m t assume  "suffix [(m, receive\<langle>t\<rangle>)] E" "prefix E A'" "A' \<in> set (tr\<^sub>p\<^sub>c A D)"
    hence "\<exists>F. prefix F A \<and> E \<in> set (tr\<^sub>p\<^sub>c F D)"
    proof (induction A D arbitrary: A' E rule: tr\<^sub>p\<^sub>c.induct)
      case (1 D) thus ?case by simp
    next
      case (2 i t' S D)
      note prems = "2.prems"
      note IH = "2.IH"
      obtain A'' where *: "A' = (i,send\<langle>t'\<rangle>)#A''" "A'' \<in> set (tr\<^sub>p\<^sub>c S D)" using prems(3) by auto
      have "E \<noteq> []" using prems(1) by auto
      then obtain E' where **: "E = (i,send\<langle>t'\<rangle>)#E'" using *(1) prems(2) by (cases E) auto
      hence "suffix [(m, receive\<langle>t\<rangle>)] E'" "prefix E' A''"
        using *(1) prems(1,2) suffix_Cons[of _ _ E'] by auto
      then obtain F where "prefix F S" "E' \<in> set (tr\<^sub>p\<^sub>c F D)" using *(2) ** IH by metis
      hence "prefix ((i,Send t')#F) ((i,Send t')#S)" "E \<in> set (tr\<^sub>p\<^sub>c ((i,Send t')#F) D)"
        using ** by auto
      thus ?case by metis
    next
      case (3 i t' S D)
      note prems = "3.prems"
      note IH = "3.IH"
      obtain A'' where *: "A' = (i,receive\<langle>t'\<rangle>)#A''" "A'' \<in> set (tr\<^sub>p\<^sub>c S D)" using prems(3) by auto
      have "E \<noteq> []" using prems(1) by auto
      then obtain E' where **: "E = (i,receive\<langle>t'\<rangle>)#E'" using *(1) prems(2) by (cases E) auto
      show ?case
      proof (cases "(m, receive\<langle>t\<rangle>) = (i, receive\<langle>t'\<rangle>)")
        case True
        note T = True
        show ?thesis
        proof (cases "suffix [(m, receive\<langle>t\<rangle>)] E'")
          case True
          hence "suffix [(m, receive\<langle>t\<rangle>)] E'" "prefix E' A''"
            using ** *(1) prems(1,2) by auto
          then obtain F where "prefix F S" "E' \<in> set (tr\<^sub>p\<^sub>c F D)" using *(2) ** IH by metis
          hence "prefix ((i,Receive t')#F) ((i,Receive t')#S)" "E \<in> set (tr\<^sub>p\<^sub>c ((i,Receive t')#F) D)"
            using ** by auto
          thus ?thesis by metis
        next
          case False
          hence "E' = []"
            using **(1) T prems(1) suffix_Cons[of "[(m, receive\<langle>t\<rangle>)]" "(m, receive\<langle>t\<rangle>)" E']
            by auto
          hence "prefix [(i,Receive t')] ((i,Receive t') # S) \<and> E \<in> set (tr\<^sub>p\<^sub>c [(i,Receive t')] D)"
            using * ** prems by auto
          thus ?thesis by metis
        qed
      next
        case False
        hence "suffix [(m, receive\<langle>t\<rangle>)] E'" "prefix E' A''"
          using ** *(1) prems(1,2) suffix_Cons[of _ _ E'] by auto
        then obtain F where "prefix F S" "E' \<in> set (tr\<^sub>p\<^sub>c F D)" using *(2) ** IH by metis
        hence "prefix ((i,Receive t')#F) ((i,Receive t')#S)" "E \<in> set (tr\<^sub>p\<^sub>c ((i,Receive t')#F) D)"
          using ** by auto
        thus ?thesis by metis
      qed
    next
      case (4 i t' t'' S D)
      note prems = "4.prems"
      note IH = "4.IH"
      obtain A'' where *: "A' = (i,\<langle>t' \<doteq> t''\<rangle>)#A''" "A'' \<in> set (tr\<^sub>p\<^sub>c S D)" using prems(3) by auto
      have "E \<noteq> []" using prems(1) by auto
      then obtain E' where **: "E = (i,\<langle>t' \<doteq> t''\<rangle>)#E'" using *(1) prems(2) by (cases E) auto
      hence "suffix [(m, receive\<langle>t\<rangle>)] E'" "prefix E' A''"
        using *(1) prems(1,2) suffix_Cons[of _ _ E'] by auto
      then obtain F where "prefix F S" "E' \<in> set (tr\<^sub>p\<^sub>c F D)" using *(2) ** IH by metis
      hence "prefix ((i,Equality t' t'')#F) ((i,Equality t' t'')#S)"
            "E \<in> set (tr\<^sub>p\<^sub>c ((i,Equality t' t'')#F) D)"
        using ** by auto
      thus ?case by metis
    next
      case (5 i X t' t'' S D)
      note prems = "5.prems"
      note IH = "5.IH"
      obtain A'' where *: "A' = (i,\<nexists>X\<langle>t' \<doteq> t''\<rangle>)#A''" "A'' \<in> set (tr\<^sub>p\<^sub>c S D)" using prems(3) by auto
      have "E \<noteq> []" using prems(1) by auto
      then obtain E' where **: "E = (i,\<nexists>X\<langle>t' \<doteq> t''\<rangle>)#E'" using *(1) prems(2) by (cases E) auto
      hence "suffix [(m, receive\<langle>t\<rangle>)] E'" "prefix E' A''"
        using *(1) prems(1,2) suffix_Cons[of _ _ E'] by auto
      then obtain F where "prefix F S" "E' \<in> set (tr\<^sub>p\<^sub>c F D)" using *(2) ** IH by metis
      hence "prefix ((i,Inequality X t' t'')#F) ((i,Inequality X t' t'')#S)"
            "E \<in> set (tr\<^sub>p\<^sub>c ((i,Inequality X t' t'')#F) D)"
        using ** by auto
      thus ?case by metis
    next
      case (6 i t' s S D)
      note prems = "6.prems"
      note IH = "6.IH"
      have *: "A' \<in> set (tr\<^sub>p\<^sub>c S (List.insert (i,t',s) D))" using prems(3) by auto
      have "E \<noteq> []" using prems(1) by auto
      hence "suffix [(m, receive\<langle>t\<rangle>)] E" "prefix E A'"
        using *(1) prems(1,2) suffix_Cons[of _ _ E] by auto
      then obtain F where "prefix F S" "E \<in> set (tr\<^sub>p\<^sub>c F (List.insert (i,t',s) D))"
        using * IH by metis
      hence "prefix ((i,Insert t' s)#F) ((i,Insert t' s)#S)"
            "E \<in> set (tr\<^sub>p\<^sub>c ((i,Insert t' s)#F) D)"
        by auto
      thus ?case by metis
    next
      case (7 i t' s S D)
      define constr where "constr = (\<lambda>Di.
        (map (\<lambda>d. (i,\<langle>to_pair (t',s) \<doteq> to_pair (snd d)\<rangle>)) Di)@
        (map (\<lambda>d. (i,\<nexists>[]\<langle>to_pair (t',s) \<doteq> to_pair (snd d)\<rangle>))
          (filter (\<lambda>d. d \<notin> set Di) (dbproj i D))))"
      note prems = "7.prems"
      note IH = "7.IH"
      obtain A'' Di where *:
          "A' = constr Di@A''" "A'' \<in> set (tr\<^sub>p\<^sub>c S (filter (\<lambda>d. d \<notin> set Di) D))"
          "Di \<in> set (subseqs (dbproj i D))"
        using prems(3) constr_def by auto
      have ***:  "(m, receive\<langle>t\<rangle>) \<notin> set (constr Di)" using constr_def by auto
      have "E \<noteq> []" using prems(1) by auto
      then obtain E' where **: "E = constr Di@E'"
        using *(1) prems(1,2) ***
        by (metis (mono_tags, lifting) Un_iff list.set_intros(1) prefixI prefix_def
                                       prefix_same_cases set_append suffix_def)
      hence "suffix [(m, receive\<langle>t\<rangle>)] E'" "prefix E' A''"
        using *(1) prems(1,2) suffix_append[of "[(m,receive\<langle>t\<rangle>)]" "constr Di" E'] ***
        by (metis (no_types, hide_lams) Nil_suffix append_Nil2 in_set_conv_decomp rev_exhaust
                                        snoc_suffix_snoc suffix_appendD,
            auto)
      then obtain F where "prefix F S" "E' \<in> set (tr\<^sub>p\<^sub>c F (filter (\<lambda>d. d \<notin> set Di) D))"
        using *(2,3) ** IH by metis
      hence "prefix ((i,Delete t' s)#F) ((i,Delete t' s)#S)"
            "E \<in> set (tr\<^sub>p\<^sub>c ((i,Delete t' s)#F) D)"
        using *(3) ** constr_def by auto
      thus ?case by metis
    next
      case (8 i t' s S D)
      define constr where "constr = (
        \<lambda>d::(('lbl strand_label \<times> ('fun,'var) term \<times> ('fun,'var) term)).
          (i,\<langle>to_pair (t',s) \<doteq> to_pair (snd d)\<rangle>))"
      note prems = "8.prems"
      note IH = "8.IH"
      obtain A'' d where *: "A' = constr d#A''" "A'' \<in> set (tr\<^sub>p\<^sub>c S D)" "d \<in> set (dbproj i D)"
        using prems(3) constr_def by auto
      have "E \<noteq> []" using prems(1) by auto
      then obtain E' where **: "E = constr d#E'" using *(1) prems(2) by (cases E) auto
      hence "suffix [(m, receive\<langle>t\<rangle>)] E'" "prefix E' A''"
        using *(1) prems(1,2) suffix_Cons[of _ _ E'] using constr_def by auto
      then obtain F where "prefix F S" "E' \<in> set (tr\<^sub>p\<^sub>c F D)" using *(2) ** IH by metis
      hence "prefix ((i,InSet t' s)#F) ((i,InSet t' s)#S)" "E \<in> set (tr\<^sub>p\<^sub>c ((i,InSet t' s)#F) D)"
        using *(3) ** unfolding constr_def by auto
      thus ?case by metis
    next
      case (9 i X t' s S D)
      define constr where "constr = map (\<lambda>d. (i,\<nexists>X\<langle>to_pair (t',s) \<doteq> to_pair (snd d)\<rangle>)) (dbproj i D)"
      note prems = "9.prems"
      note IH = "9.IH"
      obtain A'' where *: "A' = constr@A''" "A'' \<in> set (tr\<^sub>p\<^sub>c S D)"
        using prems(3) constr_def by auto
      have ***:  "(m, receive\<langle>t\<rangle>) \<notin> set constr" using constr_def by auto
      have "E \<noteq> []" using prems(1) by auto
      then obtain E' where **: "E = constr@E'"
        using *(1) prems(1,2) ***
        by (metis (mono_tags, lifting) Un_iff list.set_intros(1) prefixI prefix_def
                                       prefix_same_cases set_append suffix_def)
      hence "suffix [(m, receive\<langle>t\<rangle>)] E'" "prefix E' A''"
        using *(1) prems(1,2) suffix_append[of "[(m,receive\<langle>t\<rangle>)]" constr E'] ***
        by (metis (no_types, hide_lams) Nil_suffix append_Nil2 in_set_conv_decomp rev_exhaust
                                        snoc_suffix_snoc suffix_appendD,
            auto)
      then obtain F where "prefix F S" "E' \<in> set (tr\<^sub>p\<^sub>c F D)" using *(2) ** IH by metis
      hence "prefix ((i,NotInSet X t' s)#F) ((i,NotInSet X t' s)#S)"
            "E \<in> set (tr\<^sub>p\<^sub>c ((i,NotInSet X t' s)#F) D)"
        using ** constr_def by auto
      thus ?case by metis
    qed
  }
  moreover have "prefix [] A" "[] \<in> set (tr\<^sub>p\<^sub>c [] [])" by auto
  ultimately have 4: "\<exists>D. prefix D A \<and> C \<in> set (tr\<^sub>p\<^sub>c D [])" using C(3) assms(1) 2 by blast

  show ?thesis by (metis 1 3 4)
qed

section {* Theorem: Semantic equivalence of translation, part 1 *}
lemma tr'_par_iff_unlabel_tr:
  assumes "\<forall>(i,p) \<in> setops\<^sub>l\<^sub>s\<^sub>s\<^sub>t A \<union> set D.
           \<forall>(j,q) \<in> setops\<^sub>l\<^sub>s\<^sub>s\<^sub>t A \<union> set D.
              p = q \<longrightarrow> i = j"
shows "(\<exists>C \<in> set (tr'\<^sub>p\<^sub>c A D). B = unlabel C) \<longleftrightarrow> B \<in> set (tr (unlabel A) (unlabel D))" (is "?A \<longleftrightarrow> ?B")
proof
  { fix C have "C \<in> set (tr'\<^sub>p\<^sub>c A D) \<Longrightarrow> unlabel C \<in> set (tr (unlabel A) (unlabel D))" using assms
    proof (induction A D arbitrary: C rule: tr'\<^sub>p\<^sub>c.induct)
      case (6 i t s S D)
      hence "unlabel C \<in> set (tr (unlabel S) (unlabel (List.insert (i, t, s) D)))" by auto
      moreover have
          "insert (i,t,s) (set D) \<subseteq> setops\<^sub>l\<^sub>s\<^sub>s\<^sub>t ((i,Insert t s)#S) \<union> set D"
        by auto
      hence "\<forall>(j,p) \<in> insert (i,t,s) (set D). \<forall>(k,q) \<in> insert (i,t,s) (set D). p = q \<longrightarrow> j = k"
        using "6.prems"(2) by blast
      hence "unlabel (List.insert (i, t, s) D) = (List.insert (t, s) (unlabel D))"
        using map_snd_list_insert_distrib[of "(i,t,s)" D] unfolding unlabel_def by simp
      ultimately show ?case by auto
    next
      case (7 i t s S D)
      let ?f1 = "\<lambda>d. \<langle>to_pair (t,s) \<doteq> to_pair d\<rangle>"
      let ?g1 = "\<lambda>d. \<nexists>[]\<langle>to_pair (t,s) \<doteq> to_pair d\<rangle>"
      let ?f2 = "\<lambda>d. (fst d, ?f1 (snd d))"
      let ?g2 = "\<lambda>d. (fst d, ?g1 (snd d))"
  
      define constr1 where "constr1 = (\<lambda>Di. (map ?f1 Di)@(map ?g1 [d\<leftarrow>unlabel D. d \<notin> set Di]))"
      define constr2 where "constr2 = (\<lambda>Di. (map ?f2 Di)@(map ?g2 [d\<leftarrow>D. d \<notin> set Di]))"
  
      then obtain C' Di where C':
          "Di \<in> set (subseqs D)"
          "C = constr2 Di@C'"
          "C' \<in> set (tr'\<^sub>p\<^sub>c S [d\<leftarrow>D. d \<notin> set Di])"
        using "7.prems"(1) by moura
      
      have 0: "set [d\<leftarrow>D. d \<notin> set Di] \<subseteq> set D"
              "setops\<^sub>l\<^sub>s\<^sub>s\<^sub>t S \<subseteq> setops\<^sub>l\<^sub>s\<^sub>s\<^sub>t ((i, Delete t s)#S)"
        by auto
      hence 1:
          "\<forall>(j, p) \<in> setops\<^sub>l\<^sub>s\<^sub>s\<^sub>t S \<union> set [d\<leftarrow>D. d \<notin> set Di].
           \<forall>(k, q) \<in> setops\<^sub>l\<^sub>s\<^sub>s\<^sub>t S \<union> set [d\<leftarrow>D. d \<notin> set Di].
            p = q \<longrightarrow> j = k"
        using "7.prems"(2) by blast
  
      have "\<forall>(i,p) \<in> set D \<union> set Di. \<forall>(j,q) \<in> set D \<union> set Di. p = q \<longrightarrow> i = j"
        using "7.prems"(2) subseqs_set_subset(1)[OF C'(1)] by blast
      hence 2: "unlabel [d\<leftarrow>D. d \<notin> set Di] = [d\<leftarrow>unlabel D. d \<notin> set (unlabel Di)]"
        using unlabel_filter_eq[of D "set Di"] unfolding unlabel_def by simp
  
      have 3:
          "\<And>f g::('a \<times> 'a \<Rightarrow> 'c). \<And>A B::(('b \<times> 'a \<times> 'a) list).
              map snd ((map (\<lambda>d. (fst d, f (snd d))) A)@(map (\<lambda>d. (fst d, g (snd d))) B)) =
              map f (map snd A)@map g (map snd B)"
        by simp
      have "unlabel (constr2 Di) = constr1 (unlabel Di)"
        using 2 3[of ?f1 Di ?g1 "[d\<leftarrow>D. d \<notin> set Di]"] 
        by (simp add: constr1_def constr2_def unlabel_def)
      hence 4: "unlabel C = constr1 (unlabel Di)@unlabel C'"
        using C'(2) unlabel_append by metis
  
      have "unlabel Di \<in> set (map unlabel (subseqs D))"
        using C'(1) unfolding unlabel_def by simp
      hence 5: "unlabel Di \<in> set (subseqs (unlabel D))"
        using map_subseqs[of snd D] unfolding unlabel_def by simp
  
      show ?case using "7.IH"[OF C'(1,3) 1] 2 4 5 unfolding constr1_def by auto
    next
      case (8 i t s S D)
      then obtain C' d  where C':
          "C = (fst d,\<langle>to_pair (t,s) \<doteq> to_pair (snd d)\<rangle>)#C'"
          "C' \<in> set (tr'\<^sub>p\<^sub>c S D)" "d \<in> set D"
        by moura
  
      have "setops\<^sub>l\<^sub>s\<^sub>s\<^sub>t S \<union> set D \<subseteq> setops\<^sub>l\<^sub>s\<^sub>s\<^sub>t ((i,InSet t s)#S) \<union> set D" by auto
      hence "\<forall>(j, p) \<in> setops\<^sub>l\<^sub>s\<^sub>s\<^sub>t S \<union> set D.
             \<forall>(k, q) \<in> setops\<^sub>l\<^sub>s\<^sub>s\<^sub>t S \<union> set D.
              p = q \<longrightarrow> j = k"
        using "8.prems"(2) by blast
      hence "unlabel C' \<in> set (tr (unlabel S) (unlabel D))" using "8.IH"[OF C'(2)] by auto
      thus ?case using C' unfolding unlabel_def by force
    next
      case (9 i X t s S D)
      then obtain C' where C':
          "C = (map (\<lambda>d. (fst d,\<nexists>X\<langle>to_pair (t,s) \<doteq> to_pair (snd d)\<rangle>)) D)@C'"
          "C' \<in> set (tr'\<^sub>p\<^sub>c S D)"
        by moura
  
      have "setops\<^sub>l\<^sub>s\<^sub>s\<^sub>t S \<union> set D \<subseteq> setops\<^sub>l\<^sub>s\<^sub>s\<^sub>t ((i,NotInSet X t s)#S) \<union> set D" by auto
      hence "\<forall>(j, p) \<in> setops\<^sub>l\<^sub>s\<^sub>s\<^sub>t S \<union> set D.
             \<forall>(k, q) \<in> setops\<^sub>l\<^sub>s\<^sub>s\<^sub>t S \<union> set D.
              p = q \<longrightarrow> j = k"
        using "9.prems"(2) by blast
      hence "unlabel C' \<in> set (tr (unlabel S) (unlabel D))" using "9.IH"[OF C'(2)] by auto
      thus ?case using C' unfolding unlabel_def by auto
    qed auto
  } thus "?A \<Longrightarrow> ?B" by blast

  show "?B \<Longrightarrow> ?A" using assms
  proof (induction A arbitrary: B D)
    case (Cons a A)
    obtain ia sa where a: "a = (ia,sa)" by moura

    have "setops\<^sub>l\<^sub>s\<^sub>s\<^sub>t A \<subseteq> setops\<^sub>l\<^sub>s\<^sub>s\<^sub>t (a#A)" using a by (cases sa) auto
    hence 1: "\<forall>(j, p) \<in> setops\<^sub>l\<^sub>s\<^sub>s\<^sub>t A \<union> set D.
              \<forall>(k, q) \<in> setops\<^sub>l\<^sub>s\<^sub>s\<^sub>t A \<union> set D.
                p = q \<longrightarrow> j = k"
      using Cons.prems(2) by blast

    show ?case
    proof (cases sa)
      case (Send t)
      then obtain B' where B':
          "B = send\<langle>t\<rangle>#B'"
          "B' \<in> set (tr (unlabel A) (unlabel D))"
        using Cons.prems(1) a by auto
      thus ?thesis using Cons.IH[OF B'(2) 1] a B'(1) Send by auto
    next
      case (Receive t)
      then obtain B' where B':
          "B = receive\<langle>t\<rangle>#B'"
          "B' \<in> set (tr (unlabel A) (unlabel D))"
        using Cons.prems(1) a by auto
      thus ?thesis using Cons.IH[OF B'(2) 1] a B'(1) Receive by auto
    next
      case (Equality t t')
      then obtain B' where B':
          "B = \<langle>t \<doteq> t'\<rangle>#B'"
          "B' \<in> set (tr (unlabel A) (unlabel D))"
        using Cons.prems(1) a by auto
      thus ?thesis using Cons.IH[OF B'(2) 1] a B'(1) Equality by auto
    next
      case (Inequality X t t')
      then obtain B' where B':
          "B = \<nexists>X\<langle>t \<doteq> t'\<rangle>#B'"
          "B' \<in> set (tr (unlabel A) (unlabel D))"
        using Cons.prems(1) a by auto
      thus ?thesis using Cons.IH[OF B'(2) 1] a B'(1) Inequality by auto
    next
      case (Insert t s)
      hence B: "B \<in> set (tr (unlabel A) (List.insert (t,s) (unlabel D)))"
        using Cons.prems(1) a by auto

      let ?P = "\<lambda>i. List.insert (t,s) (unlabel D) = unlabel (List.insert (i,t,s) D)"

      { obtain j where j: "?P j" "j = ia \<or> (j,t,s) \<in> set D"
          using labeled_list_insert_eq_ex_cases[of "(t,s)" D ia] by moura
        hence "j = ia" using Cons.prems(2) a Insert by auto
        hence "?P ia" using j(1) by metis
      } hence j: "?P ia" by metis
      
      have 2: "\<forall>(k1, p) \<in> setops\<^sub>l\<^sub>s\<^sub>s\<^sub>t A \<union> set (List.insert (ia,t,s) D).
               \<forall>(k2, q) \<in> setops\<^sub>l\<^sub>s\<^sub>s\<^sub>t A \<union> set (List.insert (ia,t,s) D).
                 p = q \<longrightarrow> k1 = k2"
        using Cons.prems(2) a Insert by auto

      show ?thesis using Cons.IH[OF _ 2] j(1) B Insert a by auto
    next
      case (Delete t s)
      define c where "c \<equiv> (\<lambda>Di.
        map (\<lambda>d. (fst d,\<langle>to_pair (t,s) \<doteq> to_pair (snd d)\<rangle>)) Di@
        map (\<lambda>d. (fst d,\<nexists>[]\<langle>to_pair (t,s) \<doteq> to_pair (snd d)\<rangle>)) [d\<leftarrow>D. d \<notin> set Di])"
      
      define d where "d \<equiv> (\<lambda>Di.
        map (\<lambda>d. \<langle>to_pair (t,s) \<doteq> to_pair d\<rangle>) Di@
        map (\<lambda>d. \<nexists>[]\<langle>to_pair (t,s) \<doteq> to_pair d\<rangle>) [d\<leftarrow>unlabel D. d \<notin> set Di])"

      obtain B' Di where B':
          "B = d Di@B'" "Di \<in> set (subseqs (unlabel D))"
          "B' \<in> set (tr (unlabel A) [d\<leftarrow>unlabel D. d \<notin> set Di])"
        using Cons.prems(1) a Delete unfolding d_def by auto

      obtain Di' where Di': "Di' \<in> set (subseqs D)" "unlabel Di' = Di"
        using unlabel_subseqsD[OF B'(2)] by moura

      have 2: "\<forall>(j, p) \<in> setops\<^sub>l\<^sub>s\<^sub>s\<^sub>t A \<union> set [d\<leftarrow>D. d \<notin> set Di'].
               \<forall>(k, q) \<in> setops\<^sub>l\<^sub>s\<^sub>s\<^sub>t A \<union> set [d\<leftarrow>D. d \<notin> set Di'].
                 p = q \<longrightarrow> j = k"
        using 1 subseqs_subset[OF Di'(1)]
              filter_is_subset[of "\<lambda>d. d \<notin> set Di'"]
        by blast

      have "set Di' \<subseteq> set D" by (rule subseqs_subset[OF Di'(1)])
      hence "\<forall>(j, p)\<in>set D \<union> set Di'. \<forall>(k, q)\<in>set D \<union> set Di'. p = q \<longrightarrow> j = k"
        using Cons.prems(2) by blast
      hence 3: "[d\<leftarrow>unlabel D. d \<notin> set Di] = unlabel [d\<leftarrow>D. d \<notin> set Di']"
        using Di'(2) unlabel_filter_eq[of D "set Di'"] unfolding unlabel_def by auto
      
      obtain C where C: "C \<in> set (tr'\<^sub>p\<^sub>c A [d\<leftarrow>D. d \<notin> set Di'])" "B' = unlabel C"
        using 3 Cons.IH[OF _ 2] B'(3) by auto
      hence 4: "c Di'@C \<in> set (tr'\<^sub>p\<^sub>c (a#A) D)" using Di'(1) a Delete unfolding c_def by auto

      have "unlabel (c Di') = d Di" using Di' 3 unfolding c_def d_def unlabel_def by auto
      hence 5: "B = unlabel (c Di'@C)" using B'(1) C(2) unlabel_append[of "c Di'" C] by simp
      
      show ?thesis using 4 5 by auto
    next
      case (InSet t s)
      then obtain B' d where B':
          "B = \<langle>to_pair (t,s) \<doteq> to_pair d\<rangle>#B'"
          "B' \<in> set (tr (unlabel A) (unlabel D))"
          "d \<in> set (unlabel D)"
        using Cons.prems(1) a by auto
      thus ?thesis using Cons.IH[OF _ 1] a InSet unfolding unlabel_def by auto
    next
      case (NotInSet X t s)
      then obtain B' where B':
          "B = map (\<lambda>d. \<nexists>X\<langle>to_pair (t,s) \<doteq> to_pair d\<rangle>) (unlabel D)@B'"
          "B' \<in> set (tr (unlabel A) (unlabel D))"
        using Cons.prems(1) a by auto
      thus ?thesis using Cons.IH[OF _ 1] a NotInSet unfolding unlabel_def by auto
    qed
  qed simp
qed

section {* Theorem: Semantic equivalence of translation, part 2 *}
lemma tr_par_iff_tr'_par:
  assumes "\<forall>(i,p) \<in> setops\<^sub>l\<^sub>s\<^sub>s\<^sub>t A \<union> set D. \<forall>(j,q) \<in> setops\<^sub>l\<^sub>s\<^sub>s\<^sub>t A \<union> set D.
            (\<exists>\<delta>. Unifier \<delta> (to_pair p) (to_pair q)) \<longrightarrow> i = j" (is "?R3 A D")
  shows "(\<exists>B \<in> set (tr\<^sub>p\<^sub>c A D). \<lbrakk>M; unlabel B\<rbrakk>\<^sub>d \<I>) \<longleftrightarrow> (\<exists>C \<in> set (tr'\<^sub>p\<^sub>c A D). \<lbrakk>M; unlabel C\<rbrakk>\<^sub>d \<I>)" (is "?P \<longleftrightarrow> ?Q")
proof
  { fix B assume "B \<in> set (tr\<^sub>p\<^sub>c A D)" "\<lbrakk>M; unlabel B\<rbrakk>\<^sub>d \<I>"
    hence ?Q using assms
    proof (induction A D arbitrary: B M rule: tr\<^sub>p\<^sub>c.induct)
      case (1 D) thus ?case by simp
    next
      case (2 i t S D)
      note prems = "2.prems"
      note IH = "2.IH"

      obtain B' where B': "B = (i,send\<langle>t\<rangle>)#B'" "B' \<in> set (tr\<^sub>p\<^sub>c S D)" using prems(1) by moura
          
      have 1: "\<lbrakk>M; unlabel B'\<rbrakk>\<^sub>d \<I> " using prems(2) B'(1) by simp
      have 4: "?R3 S D" using prems(3) by auto

      have 5: "M \<cdot>\<^sub>s\<^sub>e\<^sub>t \<I> \<turnstile> t \<cdot> \<I>" using prems(2) B'(1) by simp

      obtain C where C: "C \<in> set (tr'\<^sub>p\<^sub>c S D)" "\<lbrakk>M; unlabel C\<rbrakk>\<^sub>d \<I>" using IH[OF B'(2) 1 4] by moura
      hence "((i,send\<langle>t\<rangle>)#C) \<in> set (tr'\<^sub>p\<^sub>c ((i,Send t)#S) D)" "\<lbrakk>M; unlabel ((i,send\<langle>t\<rangle>)#C)\<rbrakk>\<^sub>d \<I>"
        using 5 by auto
      thus ?case by metis
    next
      case (3 i t S D)
      note prems = "3.prems"
      note IH = "3.IH"

      obtain B' where B': "B = (i,receive\<langle>t\<rangle>)#B'" "B' \<in> set (tr\<^sub>p\<^sub>c S D)" using prems(1) by moura
          
      have 1: "\<lbrakk>insert t M; unlabel B'\<rbrakk>\<^sub>d \<I> " using prems(2) B'(1) by simp
      have 4: "?R3 S D" using prems(3) by auto

      obtain C where C: "C \<in> set (tr'\<^sub>p\<^sub>c S D)" "\<lbrakk>insert t M; unlabel C\<rbrakk>\<^sub>d \<I>"
        using IH[OF B'(2) 1 4] by moura
      hence "((i,receive\<langle>t\<rangle>)#C) \<in> set (tr'\<^sub>p\<^sub>c ((i,Receive t)#S) D)"
            "\<lbrakk>insert t M; unlabel ((i,receive\<langle>t\<rangle>)#C)\<rbrakk>\<^sub>d \<I>"
        by auto
      thus ?case by auto
    next
      case (4 i t t' S D)
      note prems = "4.prems"
      note IH = "4.IH"

      obtain B' where B': "B = (i,\<langle>t \<doteq> t'\<rangle>)#B'" "B' \<in> set (tr\<^sub>p\<^sub>c S D)" using prems(1) by moura
          
      have 1: "\<lbrakk>M; unlabel B'\<rbrakk>\<^sub>d \<I> " using prems(2) B'(1) by simp
      have 4: "?R3 S D" using prems(3) by auto

      have 5: "t \<cdot> \<I> = t' \<cdot> \<I>" using prems(2) B'(1) by simp

      obtain C where C: "C \<in> set (tr'\<^sub>p\<^sub>c S D)" "\<lbrakk>M; unlabel C\<rbrakk>\<^sub>d \<I>" using IH[OF B'(2) 1 4] by moura
      hence "((i,\<langle>t \<doteq> t'\<rangle>)#C) \<in> set (tr'\<^sub>p\<^sub>c ((i,Equality t t')#S) D)"
            "\<lbrakk>M; unlabel ((i,\<langle>t \<doteq> t'\<rangle>)#C)\<rbrakk>\<^sub>d \<I>"
        using 5 by auto
      thus ?case by metis
    next
      case (5 i X t t' S D)
      note prems = "5.prems"
      note IH = "5.IH"

      obtain B' where B': "B = (i,\<nexists>X\<langle>t \<doteq> t'\<rangle>)#B'" "B' \<in> set (tr\<^sub>p\<^sub>c S D)" using prems(1) by moura
          
      have 1: "\<lbrakk>M; unlabel B'\<rbrakk>\<^sub>d \<I> " using prems(2) B'(1) by simp
      have 4: "?R3 S D" using prems(3) by auto

      have 5: "ineq_model \<I> X t t'" using prems(2) B'(1) by simp

      obtain C where C: "C \<in> set (tr'\<^sub>p\<^sub>c S D)" "\<lbrakk>M; unlabel C\<rbrakk>\<^sub>d \<I>" using IH[OF B'(2) 1 4] by moura
      hence "((i,\<nexists>X\<langle>t \<doteq> t'\<rangle>)#C) \<in> set (tr'\<^sub>p\<^sub>c ((i,Inequality X t t')#S) D)"
            "\<lbrakk>M; unlabel ((i,\<nexists>X\<langle>t \<doteq> t'\<rangle>)#C)\<rbrakk>\<^sub>d \<I>"
        using 5 by auto
      thus ?case by metis
    next
      case (6 i t s S D)
      note prems = "6.prems"
      note IH = "6.IH"

      have B: "B \<in> set (tr\<^sub>p\<^sub>c S (List.insert (i,t,s) D))" using prems(1) by simp
          
      have 1: "\<lbrakk>M; unlabel B\<rbrakk>\<^sub>d \<I> " using prems(2) B(1) by simp
      have 4: "?R3 S (List.insert (i,t,s) D)" using prems(3) by auto

      show ?case using IH[OF B(1) 1 4] by simp
    next
      case (7 i t s S D)
      note prems = "7.prems"
      note IH = "7.IH"

      let ?cl1 = "\<lambda>Di. map (\<lambda>d. (i,\<langle>to_pair (t,s) \<doteq> to_pair (snd d)\<rangle>)) Di"
      let ?cu1 = "\<lambda>Di. map (\<lambda>d. \<langle>to_pair (t,s) \<doteq> to_pair (snd d)\<rangle>) Di"
      let ?cl2 = "\<lambda>Di. map (\<lambda>d. (i,\<nexists>[]\<langle>to_pair (t,s) \<doteq> to_pair (snd d)\<rangle>)) [d\<leftarrow>dbproj i D. d\<notin>set Di]"
      let ?cu2 = "\<lambda>Di. map (\<lambda>d. \<nexists>[]\<langle>to_pair (t,s) \<doteq> to_pair (snd d)\<rangle>) [d\<leftarrow>dbproj i D. d\<notin>set Di]"

      let ?dl1 = "\<lambda>Di. map (\<lambda>d. (fst d,\<langle>to_pair (t,s) \<doteq> to_pair (snd d)\<rangle>)) Di"
      let ?du1 = "\<lambda>Di. map (\<lambda>d. \<langle>to_pair (t,s) \<doteq> to_pair (snd d)\<rangle>) Di"
      let ?dl2 = "\<lambda>Di. map (\<lambda>d. (fst d,\<nexists>[]\<langle>to_pair (t,s) \<doteq> to_pair (snd d)\<rangle>)) [d\<leftarrow>D. d\<notin>set Di]"
      let ?du2 = "\<lambda>Di. map (\<lambda>d. \<nexists>[]\<langle>to_pair (t,s) \<doteq> to_pair (snd d)\<rangle>) [d\<leftarrow>D. d\<notin>set Di]"

      define c where c: "c = (\<lambda>Di. ?cl1 Di@?cl2 Di)"
      define d where d: "d = (\<lambda>Di. ?dl1 Di@?dl2 Di)"

      obtain B' Di where B':
          "Di \<in> set (subseqs (dbproj i D))" "B = c Di@B'" "B' \<in> set (tr\<^sub>p\<^sub>c S [d\<leftarrow>D. d \<notin> set Di])"
        using prems(1) c by moura

      have 0: "ik\<^sub>s\<^sub>t (unlabel (c Di)) = {}" "ik\<^sub>s\<^sub>t (unlabel (d Di)) = {}"
              "unlabel (?cl1 Di) = ?cu1 Di" "unlabel (?cl2 Di) = ?cu2 Di"
              "unlabel (?dl1 Di) = ?du1 Di" "unlabel (?dl2 Di) = ?du2 Di"
        unfolding c d unlabel_def by force+

      have 1: "\<lbrakk>M; unlabel B'\<rbrakk>\<^sub>d \<I> " using prems(2) B'(2) 0(1) unfolding unlabel_def by auto

      { fix j p k q
        assume "(j, p) \<in> setops\<^sub>l\<^sub>s\<^sub>s\<^sub>t S \<union> set [d\<leftarrow>D. d \<notin> set Di]"
               "(k, q) \<in> setops\<^sub>l\<^sub>s\<^sub>s\<^sub>t S \<union> set [d\<leftarrow>D. d \<notin> set Di]"
        hence "(j, p) \<in> setops\<^sub>l\<^sub>s\<^sub>s\<^sub>t ((i, Delete t s)#S) \<union> set D"
              "(k, q) \<in> setops\<^sub>l\<^sub>s\<^sub>s\<^sub>t ((i, Delete t s)#S) \<union> set D"
          using dbproj_subseq_subset[OF B'(1)] by auto
        hence "(\<exists>\<delta>. Unifier \<delta> (to_pair p) (to_pair q)) \<Longrightarrow> j = k" using prems(3) by blast
      } hence 4: "?R3 S [d\<leftarrow>D. d \<notin> set Di]" by blast
      
      have 5: "\<lbrakk>M; unlabel (c Di)\<rbrakk>\<^sub>d \<I>" "\<lbrakk>M; unlabel B'\<rbrakk>\<^sub>d \<I>"
        using prems(2) B'(2) 0(1)
              strand_sem_split_left_d[of M "unlabel (c Di)" "unlabel B'"]
              strand_sem_split_right_d[of M "unlabel (c Di)" "unlabel B'"]
        unfolding c unlabel_def by auto

      obtain C where C: "C \<in> set (tr'\<^sub>p\<^sub>c S [d\<leftarrow>D . d \<notin> set Di])" "\<lbrakk>M; unlabel C\<rbrakk>\<^sub>d \<I>"
        using IH[OF B'(1,3) 1 4] by moura

      have 6: "d Di@C \<in> set (tr'\<^sub>p\<^sub>c ((i,Delete t s)#S) D)"
        using C(1) dbproj_subseq_in_subseqs[OF B'(1)]
        unfolding d unlabel_def by auto

      have "\<lbrakk>M; unlabel (?cl2 Di)\<rbrakk>\<^sub>d \<I>" using 5(1) 0(1) unfolding c unlabel_def by auto 
      hence "\<lbrakk>M; ?cu2 Di\<rbrakk>\<^sub>d \<I>" by (metis 0(4))
      moreover {
        fix j p k q
        assume "(j, p) \<in> {(i, t, s)} \<union> set D \<union> set Di"
               "(k, q) \<in> {(i, t, s)} \<union> set D \<union> set Di"
        hence "(j, p) \<in> setops\<^sub>l\<^sub>s\<^sub>s\<^sub>t ((i, Delete t s)#S) \<union> set D"
              "(k, q) \<in> setops\<^sub>l\<^sub>s\<^sub>s\<^sub>t ((i, Delete t s)#S) \<union> set D"
          using dbproj_subseq_subset[OF B'(1)] by auto
        hence "(\<exists>\<delta>. Unifier \<delta> (to_pair p) (to_pair q)) \<Longrightarrow> j = k" using prems(3) by blast
      } hence "\<forall>(j, p) \<in> {(i, t, s)} \<union> set D \<union> set Di.
               \<forall>(k, q) \<in> {(i, t, s)} \<union> set D \<union> set Di.
                (\<exists>\<delta>. Unifier \<delta> (to_pair p) (to_pair q)) \<longrightarrow> j = k"
        by blast
      ultimately have "\<lbrakk>M; ?du2 Di\<rbrakk>\<^sub>d \<I>" using labeled_sat_ineq_lift by simp
      hence "\<lbrakk>M; unlabel (?dl2 Di)\<rbrakk>\<^sub>d \<I>" by (metis 0(6))
      moreover have "\<lbrakk>M; unlabel (?cl1 Di)\<rbrakk>\<^sub>d \<I>" using 5(1) unfolding c unlabel_def by auto 
      hence "\<lbrakk>M; unlabel (?dl1 Di)\<rbrakk>\<^sub>d \<I>" by (metis 0(3,5))
      ultimately have "\<lbrakk>M; unlabel (d Di)\<rbrakk>\<^sub>d \<I>" using 0(2) unfolding c d unlabel_def by force
      hence 7: "\<lbrakk>M; unlabel (d Di@C)\<rbrakk>\<^sub>d \<I>" using 0(2) C(2) unfolding unlabel_def by auto

      show ?case by (metis 6 7)
    next
      case (8 i t s S D)
      note prems = "8.prems"
      note IH = "8.IH"

      obtain B' d where B':
          "B = (i,\<langle>to_pair (t,s) \<doteq> to_pair (snd d)\<rangle>)#B'"
          "B' \<in> set (tr\<^sub>p\<^sub>c S D)" "d \<in> set (dbproj i D)"
        using prems(1) by moura
          
      have 1: "\<lbrakk>M; unlabel B'\<rbrakk>\<^sub>d \<I> " using prems(2) B'(1) by simp

      { fix j p k q
        assume "(j,p) \<in> setops\<^sub>l\<^sub>s\<^sub>s\<^sub>t S \<union> set D"
               "(k,q) \<in> setops\<^sub>l\<^sub>s\<^sub>s\<^sub>t S \<union> set D"
        hence "(j,p) \<in> setops\<^sub>l\<^sub>s\<^sub>s\<^sub>t ((i, InSet t s)#S) \<union> set D"
              "(k,q) \<in> setops\<^sub>l\<^sub>s\<^sub>s\<^sub>t ((i, InSet t s)#S) \<union> set D"
          by auto
        hence "(\<exists>\<delta>. Unifier \<delta> (to_pair p) (to_pair q)) \<Longrightarrow> j = k" using prems(3) by blast
      } hence 4: "?R3 S D" by blast

      have 5: "to_pair (t,s) \<cdot> \<I> = to_pair (snd d) \<cdot> \<I>" using prems(2) B'(1) by simp

      obtain C where C: "C \<in> set (tr'\<^sub>p\<^sub>c S D)" "\<lbrakk>M; unlabel C\<rbrakk>\<^sub>d \<I>" using IH[OF B'(2) 1 4] by moura
      hence "((i,\<langle>to_pair (t,s) \<doteq> to_pair (snd d)\<rangle>)#C) \<in> set (tr'\<^sub>p\<^sub>c ((i,InSet t s)#S) D)"
            "\<lbrakk>M; unlabel ((i,\<langle>to_pair (t,s) \<doteq> to_pair (snd d)\<rangle>)#C)\<rbrakk>\<^sub>d \<I>"
        using 5 B'(3) by auto
      thus ?case by metis
    next
      case (9 i X t s S D)
      note prems = "9.prems"
      note IH = "9.IH"

      let ?cl = "map (\<lambda>d. (i,\<nexists>X\<langle>to_pair (t,s) \<doteq> to_pair (snd d)\<rangle>)) (dbproj i D)"
      let ?cu = "map (\<lambda>d. \<nexists>X\<langle>to_pair (t,s) \<doteq> to_pair (snd d)\<rangle>) (dbproj i D)"

      let ?dl = "map (\<lambda>d. (fst d,\<nexists>X\<langle>to_pair (t,s) \<doteq> to_pair (snd d)\<rangle>)) D"
      let ?du = "map (\<lambda>d. \<nexists>X\<langle>to_pair (t,s) \<doteq> to_pair (snd d)\<rangle>) D"

      define c where c: "c = ?cl"
      define d where d: "d = ?dl"

      obtain B' where B': "B = c@B'" "B' \<in> set (tr\<^sub>p\<^sub>c S D)" using prems(1) c by moura

      have 0: "ik\<^sub>s\<^sub>t (unlabel c) = {}" "ik\<^sub>s\<^sub>t (unlabel d) = {}"
              "unlabel ?cl = ?cu" "unlabel ?dl = ?du" 
        unfolding c d unlabel_def by force+

      have "ik\<^sub>s\<^sub>t (unlabel c) = {}" unfolding c unlabel_def by force
      hence 1: "\<lbrakk>M; unlabel B'\<rbrakk>\<^sub>d \<I> " using prems(2) B'(1) unfolding unlabel_def by auto

      have "setops\<^sub>l\<^sub>s\<^sub>s\<^sub>t S \<subseteq> setops\<^sub>l\<^sub>s\<^sub>s\<^sub>t ((i, NotInSet X t s)#S)" by auto
      hence 4: "?R3 S D" using prems(3) by blast

      have 5: "\<lbrakk>M; unlabel c\<rbrakk>\<^sub>d \<I>" "\<lbrakk>M; unlabel B'\<rbrakk>\<^sub>d \<I>"
        using prems(2) B'(1) 0(1)
              strand_sem_split_left_d[of M "unlabel c" "unlabel B'"]
              strand_sem_split_right_d[of M "unlabel c" "unlabel B'"]
        unfolding c unlabel_def by auto

      obtain C where C: "C \<in> set (tr'\<^sub>p\<^sub>c S D)" "\<lbrakk>M; unlabel C\<rbrakk>\<^sub>d \<I>" using IH[OF B'(2) 1 4] by moura

      have 6: "d@C \<in> set (tr'\<^sub>p\<^sub>c ((i,NotInSet X t s)#S) D)" using C(1) unfolding d unlabel_def by auto

      have "\<lbrakk>M; unlabel ?cl\<rbrakk>\<^sub>d \<I>" using 5(1) unfolding c unlabel_def by auto 
      hence "\<lbrakk>M; ?cu\<rbrakk>\<^sub>d \<I>" by (metis 0(3))
      moreover {
        fix j p k q
        assume "(j, p) \<in> {(i, t, s)} \<union> set D"
               "(k, q) \<in> {(i, t, s)} \<union> set D"
        hence "(j, p) \<in> setops\<^sub>l\<^sub>s\<^sub>s\<^sub>t ((i, NotInSet X t s)#S) \<union> set D"
              "(k, q) \<in> setops\<^sub>l\<^sub>s\<^sub>s\<^sub>t ((i, NotInSet X t s)#S) \<union> set D"
          by auto
        hence "(\<exists>\<delta>. Unifier \<delta> (to_pair p) (to_pair q)) \<Longrightarrow> j = k" using prems(3) by blast
      } hence "\<forall>(j, p) \<in> {(i, t, s)} \<union> set D. \<forall>(k, q) \<in> {(i, t, s)} \<union> set D.
                (\<exists>\<delta>. Unifier \<delta> (to_pair p) (to_pair q)) \<longrightarrow> j = k"
        by blast
      ultimately have "\<lbrakk>M; ?du\<rbrakk>\<^sub>d \<I>" using labeled_sat_ineq_lift' by simp
      hence "\<lbrakk>M; unlabel ?dl\<rbrakk>\<^sub>d \<I>" by (metis 0(4))
      hence "\<lbrakk>M; unlabel d\<rbrakk>\<^sub>d \<I>" using 0(2) unfolding c d unlabel_def by force
      hence 7: "\<lbrakk>M; unlabel (d@C)\<rbrakk>\<^sub>d \<I>" using 0(2) C(2) unfolding unlabel_def by auto

      show ?case by (metis 6 7)
    qed
  } thus "?P \<Longrightarrow> ?Q" by metis

  { fix C assume "C \<in> set (tr'\<^sub>p\<^sub>c A D)" "\<lbrakk>M; unlabel C\<rbrakk>\<^sub>d \<I>"
    hence ?P using assms
    proof (induction A D arbitrary: C M rule: tr'\<^sub>p\<^sub>c.induct)
      case (1 D) thus ?case by simp
    next
      case (2 i t S D)
      note prems = "2.prems"
      note IH = "2.IH"

      obtain C' where C': "C = (i,send\<langle>t\<rangle>)#C'" "C' \<in> set (tr'\<^sub>p\<^sub>c S D)" using prems(1) by moura
          
      have 1: "\<lbrakk>M; unlabel C'\<rbrakk>\<^sub>d \<I> " using prems(2) C'(1) by simp
      have 4: "?R3 S D" using prems(3) by auto

      have 5: "M \<cdot>\<^sub>s\<^sub>e\<^sub>t \<I> \<turnstile> t \<cdot> \<I>" using prems(2) C'(1) by simp

      obtain B where B: "B \<in> set (tr\<^sub>p\<^sub>c S D)" "\<lbrakk>M; unlabel B\<rbrakk>\<^sub>d \<I>" using IH[OF C'(2) 1 4] by moura
      hence "((i,send\<langle>t\<rangle>)#B) \<in> set (tr\<^sub>p\<^sub>c ((i,Send t)#S) D)" "\<lbrakk>M; unlabel ((i,send\<langle>t\<rangle>)#B)\<rbrakk>\<^sub>d \<I>"
        using 5 by auto
      thus ?case by metis
    next
      case (3 i t S D)
      note prems = "3.prems"
      note IH = "3.IH"

      obtain C' where C': "C = (i,receive\<langle>t\<rangle>)#C'" "C' \<in> set (tr'\<^sub>p\<^sub>c S D)" using prems(1) by moura
          
      have 1: "\<lbrakk>insert t M; unlabel C'\<rbrakk>\<^sub>d \<I> " using prems(2) C'(1) by simp
      have 4: "?R3 S D" using prems(3) by auto

      obtain B where B: "B \<in> set (tr\<^sub>p\<^sub>c S D)" "\<lbrakk>insert t M; unlabel B\<rbrakk>\<^sub>d \<I>"
        using IH[OF C'(2) 1 4] by moura
      hence "((i,receive\<langle>t\<rangle>)#B) \<in> set (tr\<^sub>p\<^sub>c ((i,Receive t)#S) D)"
            "\<lbrakk>insert t M; unlabel ((i,receive\<langle>t\<rangle>)#B)\<rbrakk>\<^sub>d \<I>"
        by auto
      thus ?case by auto
    next
      case (4 i t t' S D)
      note prems = "4.prems"
      note IH = "4.IH"

      obtain C' where C': "C = (i,\<langle>t \<doteq> t'\<rangle>)#C'" "C' \<in> set (tr'\<^sub>p\<^sub>c S D)" using prems(1) by moura
          
      have 1: "\<lbrakk>M; unlabel C'\<rbrakk>\<^sub>d \<I> " using prems(2) C'(1) by simp
      have 4: "?R3 S D" using prems(3) by auto

      have 5: "t \<cdot> \<I> = t' \<cdot> \<I>" using prems(2) C'(1) by simp

      obtain B where B: "B \<in> set (tr\<^sub>p\<^sub>c S D)" "\<lbrakk>M; unlabel B\<rbrakk>\<^sub>d \<I>" using IH[OF C'(2) 1 4] by moura
      hence "((i,\<langle>t \<doteq> t'\<rangle>)#B) \<in> set (tr\<^sub>p\<^sub>c ((i,Equality t t')#S) D)"
            "\<lbrakk>M; unlabel ((i,\<langle>t \<doteq> t'\<rangle>)#B)\<rbrakk>\<^sub>d \<I>"
        using 5 by auto
      thus ?case by metis
    next
      case (5 i X t t' S D)
      note prems = "5.prems"
      note IH = "5.IH"

      obtain C' where C': "C = (i,\<nexists>X\<langle>t \<doteq> t'\<rangle>)#C'" "C' \<in> set (tr'\<^sub>p\<^sub>c S D)" using prems(1) by moura
          
      have 1: "\<lbrakk>M; unlabel C'\<rbrakk>\<^sub>d \<I> " using prems(2) C'(1) by simp
      have 4: "?R3 S D" using prems(3) by auto

      have 5: "ineq_model \<I> X t t'" using prems(2) C'(1) by simp

      obtain B where B: "B \<in> set (tr\<^sub>p\<^sub>c S D)" "\<lbrakk>M; unlabel B\<rbrakk>\<^sub>d \<I>" using IH[OF C'(2) 1 4] by moura
      hence "((i,\<nexists>X\<langle>t \<doteq> t'\<rangle>)#B) \<in> set (tr\<^sub>p\<^sub>c ((i,Inequality X t t')#S) D)"
            "\<lbrakk>M; unlabel ((i,\<nexists>X\<langle>t \<doteq> t'\<rangle>)#B)\<rbrakk>\<^sub>d \<I>"
        using 5 by auto
      thus ?case by metis
    next
      case (6 i t s S D)
      note prems = "6.prems"
      note IH = "6.IH"

      have C: "C \<in> set (tr'\<^sub>p\<^sub>c S (List.insert (i,t,s) D))" using prems(1) by simp
          
      have 1: "\<lbrakk>M; unlabel C\<rbrakk>\<^sub>d \<I> " using prems(2) C(1) by simp
      have 4: "?R3 S (List.insert (i,t,s) D)" using prems(3) by auto

      show ?case using IH[OF C(1) 1 4] by simp
    next
      case (7 i t s S D)
      note prems = "7.prems"
      note IH = "7.IH"

      let ?dl1 = "\<lambda>Di. map (\<lambda>d. (i,\<langle>to_pair (t,s) \<doteq> to_pair (snd d)\<rangle>)) Di"
      let ?du1 = "\<lambda>Di. map (\<lambda>d. \<langle>to_pair (t,s) \<doteq> to_pair (snd d)\<rangle>) Di"
      let ?dl2 = "\<lambda>Di. map (\<lambda>d. (i,\<nexists>[]\<langle>to_pair (t,s) \<doteq> to_pair (snd d)\<rangle>)) [d\<leftarrow>dbproj i D. d\<notin>set Di]"
      let ?du2 = "\<lambda>Di. map (\<lambda>d. \<nexists>[]\<langle>to_pair (t,s) \<doteq> to_pair (snd d)\<rangle>) [d\<leftarrow>dbproj i D. d\<notin>set Di]"

      let ?cl1 = "\<lambda>Di. map (\<lambda>d. (fst d,\<langle>to_pair (t,s) \<doteq> to_pair (snd d)\<rangle>)) Di"
      let ?cu1 = "\<lambda>Di. map (\<lambda>d. \<langle>to_pair (t,s) \<doteq> to_pair (snd d)\<rangle>) Di"
      let ?cl2 = "\<lambda>Di. map (\<lambda>d. (fst d,\<nexists>[]\<langle>to_pair (t,s) \<doteq> to_pair (snd d)\<rangle>)) [d\<leftarrow>D. d\<notin>set Di]"
      let ?cu2 = "\<lambda>Di. map (\<lambda>d. \<nexists>[]\<langle>to_pair (t,s) \<doteq> to_pair (snd d)\<rangle>) [d\<leftarrow>D. d\<notin>set Di]"

      define c where c: "c = (\<lambda>Di. ?cl1 Di@?cl2 Di)"
      define d where d: "d = (\<lambda>Di. ?dl1 Di@?dl2 Di)"

      obtain C' Di where C':
          "Di \<in> set (subseqs D)" "C = c Di@C'" "C' \<in> set (tr'\<^sub>p\<^sub>c S [d\<leftarrow>D. d \<notin> set Di])"
        using prems(1) c by moura

      have 0: "ik\<^sub>s\<^sub>t (unlabel (c Di)) = {}" "ik\<^sub>s\<^sub>t (unlabel (d Di)) = {}"
              "unlabel (?cl1 Di) = ?cu1 Di" "unlabel (?cl2 Di) = ?cu2 Di"
              "unlabel (?dl1 Di) = ?du1 Di" "unlabel (?dl2 Di) = ?du2 Di"
        unfolding c d unlabel_def by force+

      have 1: "\<lbrakk>M; unlabel C'\<rbrakk>\<^sub>d \<I> " using prems(2) C'(2) 0(1) unfolding unlabel_def by auto

      { fix j p k q
        assume "(j, p) \<in> setops\<^sub>l\<^sub>s\<^sub>s\<^sub>t S \<union> set [d\<leftarrow>D. d \<notin> set Di]"
               "(k, q) \<in> setops\<^sub>l\<^sub>s\<^sub>s\<^sub>t S \<union> set [d\<leftarrow>D. d \<notin> set Di]"
        hence "(j, p) \<in> setops\<^sub>l\<^sub>s\<^sub>s\<^sub>t ((i, Delete t s)#S) \<union> set D"
              "(k, q) \<in> setops\<^sub>l\<^sub>s\<^sub>s\<^sub>t ((i, Delete t s)#S) \<union> set D"
          by auto
        hence "(\<exists>\<delta>. Unifier \<delta> (to_pair p) (to_pair q)) \<Longrightarrow> j = k" using prems(3) by blast
      } hence 4: "?R3 S [d\<leftarrow>D. d \<notin> set Di]" by blast
      
      have 5: "\<lbrakk>M; unlabel (c Di)\<rbrakk>\<^sub>d \<I>" "\<lbrakk>M; unlabel C'\<rbrakk>\<^sub>d \<I>"
        using prems(2) C'(2) 0(1)
              strand_sem_split_left_d[of M "unlabel (c Di)" "unlabel C'"]
              strand_sem_split_right_d[of M "unlabel (c Di)" "unlabel C'"]
        unfolding c unlabel_def by auto

      obtain B where B: "B \<in> set (tr\<^sub>p\<^sub>c S [d\<leftarrow>D. d \<notin> set Di])" "\<lbrakk>M; unlabel B\<rbrakk>\<^sub>d \<I>"
        using IH[OF C'(1,3) 1 4] by moura

      { fix j p k q
        assume "(j, p) \<in> {(i, t, s)} \<union> set D"
               "(k, q) \<in> {(i, t, s)} \<union> set D"
        hence "(j, p) \<in> setops\<^sub>l\<^sub>s\<^sub>s\<^sub>t ((i, Delete t s)#S) \<union> set D"
              "(k, q) \<in> setops\<^sub>l\<^sub>s\<^sub>s\<^sub>t ((i, Delete t s)#S) \<union> set D"
          by auto
        hence "(\<exists>\<delta>. Unifier \<delta> (to_pair p) (to_pair q)) \<Longrightarrow> j = k" using prems(3) by blast
      } hence "\<forall>(j, p) \<in> {(i, t, s)} \<union> set D.
               \<forall>(k, q) \<in> {(i, t, s)} \<union> set D.
                (\<exists>\<delta>. Unifier \<delta> (to_pair p) (to_pair q)) \<longrightarrow> j = k"
        by blast
      moreover have "\<lbrakk>M; unlabel (?cl1 Di)\<rbrakk>\<^sub>d \<I>" using 5(1) unfolding c unlabel_append by auto 
      hence "\<lbrakk>M; ?cu1 Di\<rbrakk>\<^sub>d \<I>" by (metis 0(3))
      ultimately have *: "Di \<in> set (subseqs (dbproj i D))"
        using labeled_sat_eqs_subseqs[OF C'(1)] by simp
      hence 6: "d Di@B \<in> set (tr\<^sub>p\<^sub>c ((i,Delete t s)#S) D)"
        using B(1) unfolding d unlabel_def by auto

      have "\<lbrakk>M; unlabel (?cl2 Di)\<rbrakk>\<^sub>d \<I>" using 5(1) 0(1) unfolding c unlabel_def by auto 
      hence "\<lbrakk>M; ?cu2 Di\<rbrakk>\<^sub>d \<I>" by (metis 0(4))
      hence "\<lbrakk>M; ?du2 Di\<rbrakk>\<^sub>d \<I>" by (metis labeled_sat_ineq_dbproj)
      hence "\<lbrakk>M; unlabel (?dl2 Di)\<rbrakk>\<^sub>d \<I>" by (metis 0(6))
      moreover have "\<lbrakk>M; unlabel (?cl1 Di)\<rbrakk>\<^sub>d \<I>" using 5(1) unfolding c unlabel_def by auto 
      hence "\<lbrakk>M; unlabel (?dl1 Di)\<rbrakk>\<^sub>d \<I>" by (metis 0(3,5))
      ultimately have "\<lbrakk>M; unlabel (d Di)\<rbrakk>\<^sub>d \<I>" using 0(2) unfolding c d unlabel_def by force
      hence 7: "\<lbrakk>M; unlabel (d Di@B)\<rbrakk>\<^sub>d \<I>" using 0(2) B(2) unfolding unlabel_def by auto

      show ?case by (metis 6 7)
    next
      case (8 i t s S D)
      note prems = "8.prems"
      note IH = "8.IH"

      obtain C' d where C':
          "C = (fst d,\<langle>to_pair (t,s) \<doteq> to_pair (snd d)\<rangle>)#C'" "C' \<in> set (tr'\<^sub>p\<^sub>c S D)" "d \<in> set D"
        using prems(1) by moura
          
      have 1: "\<lbrakk>M; unlabel C'\<rbrakk>\<^sub>d \<I> " using prems(2) C'(1) by simp

      { fix j p k q
        assume "(j,p) \<in> setops\<^sub>l\<^sub>s\<^sub>s\<^sub>t S \<union> set D"
               "(k,q) \<in> setops\<^sub>l\<^sub>s\<^sub>s\<^sub>t S \<union> set D"
        hence "(j,p) \<in> setops\<^sub>l\<^sub>s\<^sub>s\<^sub>t ((i, InSet t s)#S) \<union> set D"
              "(k,q) \<in> setops\<^sub>l\<^sub>s\<^sub>s\<^sub>t ((i, InSet t s)#S) \<union> set D"
          by auto
        hence "(\<exists>\<delta>. Unifier \<delta> (to_pair p) (to_pair q)) \<Longrightarrow> j = k" using prems(3) by blast
      } hence 4: "?R3 S D" by blast

      have 5: "to_pair (t,s) \<cdot> \<I> = to_pair (snd d) \<cdot> \<I>" using prems(2) C'(1) by simp

      have "(i,t,s) \<in> setops\<^sub>l\<^sub>s\<^sub>s\<^sub>t ((i, InSet t s)#S) \<union> set D"
           "(fst d, snd d) \<in> setops\<^sub>l\<^sub>s\<^sub>s\<^sub>t ((i, InSet t s)#S) \<union> set D"
        using C'(3) by auto
      hence "\<exists>\<delta>. Unifier \<delta> (to_pair (t,s)) (to_pair (snd d)) \<Longrightarrow> i = fst d"
        using prems(3) by blast
      hence "fst d = i" using 5 by auto
      hence 6: "d \<in> set (dbproj i D)" using C'(3) by auto

      obtain B where B: "B \<in> set (tr\<^sub>p\<^sub>c S D)" "\<lbrakk>M; unlabel B\<rbrakk>\<^sub>d \<I>" using IH[OF C'(2) 1 4] by moura
      have 7: "((i,\<langle>to_pair (t,s) \<doteq> to_pair (snd d)\<rangle>)#B) \<in> set (tr\<^sub>p\<^sub>c ((i,InSet t s)#S) D)"
        using B 6 by auto
      have 8: "\<lbrakk>M; unlabel ((i,\<langle>to_pair (t,s) \<doteq> to_pair (snd d)\<rangle>)#B)\<rbrakk>\<^sub>d \<I>"
        using B 5 6 by auto

      show ?case by (metis 7 8)
    next
      case (9 i X t s S D)
      note prems = "9.prems"
      note IH = "9.IH"

      let ?dl = "map (\<lambda>d. (i,\<nexists>X\<langle>to_pair (t,s) \<doteq> to_pair (snd d)\<rangle>)) (dbproj i D)"
      let ?du = "map (\<lambda>d. \<nexists>X\<langle>to_pair (t,s) \<doteq> to_pair (snd d)\<rangle>) (dbproj i D)"

      let ?cl = "map (\<lambda>d. (fst d,\<nexists>X\<langle>to_pair (t,s) \<doteq> to_pair (snd d)\<rangle>)) D"
      let ?cu = "map (\<lambda>d. \<nexists>X\<langle>to_pair (t,s) \<doteq> to_pair (snd d)\<rangle>) D"

      define c where c: "c = ?cl"
      define d where d: "d = ?dl"

      obtain C' where C': "C = c@C'" "C' \<in> set (tr'\<^sub>p\<^sub>c S D)" using prems(1) c by moura

      have 0: "ik\<^sub>s\<^sub>t (unlabel c) = {}" "ik\<^sub>s\<^sub>t (unlabel d) = {}"
              "unlabel ?cl = ?cu" "unlabel ?dl = ?du" 
        unfolding c d unlabel_def by force+

      have "ik\<^sub>s\<^sub>t (unlabel c) = {}" unfolding c unlabel_def by force
      hence 1: "\<lbrakk>M; unlabel C'\<rbrakk>\<^sub>d \<I> " using prems(2) C'(1) unfolding unlabel_def by auto

      have "setops\<^sub>l\<^sub>s\<^sub>s\<^sub>t S \<subseteq> setops\<^sub>l\<^sub>s\<^sub>s\<^sub>t ((i, NotInSet X t s)#S)" by auto
      hence 4: "?R3 S D" using prems(3) by blast

      have 5: "\<lbrakk>M; unlabel c\<rbrakk>\<^sub>d \<I>" "\<lbrakk>M; unlabel C'\<rbrakk>\<^sub>d \<I>"
        using prems(2) C'(1) 0(1)
              strand_sem_split_left_d[of M "unlabel c" "unlabel C'"]
              strand_sem_split_right_d[of M "unlabel c" "unlabel C'"]
        unfolding c unlabel_def by auto

      obtain B where B: "B \<in> set (tr\<^sub>p\<^sub>c S D)" "\<lbrakk>M; unlabel B\<rbrakk>\<^sub>d \<I>" using IH[OF C'(2) 1 4] by moura

      have 6: "d@B \<in> set (tr\<^sub>p\<^sub>c ((i,NotInSet X t s)#S) D)" using B(1) unfolding d unlabel_def by auto

      have "\<lbrakk>M; unlabel ?cl\<rbrakk>\<^sub>d \<I>" using 5(1) unfolding c unlabel_def by auto 
      hence "\<lbrakk>M; ?cu\<rbrakk>\<^sub>d \<I>" by (metis 0(3))
      hence "\<lbrakk>M; ?du\<rbrakk>\<^sub>d \<I>" by (metis labeled_sat_ineq_dbproj')
      hence "\<lbrakk>M; unlabel ?dl\<rbrakk>\<^sub>d \<I>" by (metis 0(4))
      hence "\<lbrakk>M; unlabel d\<rbrakk>\<^sub>d \<I>" using 0(2) unfolding c d unlabel_def by force
      hence 7: "\<lbrakk>M; unlabel (d@B)\<rbrakk>\<^sub>d \<I>" using 0(2) B(2) unfolding unlabel_def by auto

      show ?case by (metis 6 7)
    qed
  } thus "?Q \<Longrightarrow> ?P" by metis
qed


section {* Theorem: Semantic equivalence of translation, part 3 *}
lemma tr'_par_sem_equiv:
  assumes "\<forall>(l,t,s) \<in> set D. (fv t \<union> fv s) \<inter> bvars\<^sub>s\<^sub>s\<^sub>t (unlabel A) = {}"
  and "fv\<^sub>s\<^sub>s\<^sub>t (unlabel A) \<inter> bvars\<^sub>s\<^sub>s\<^sub>t (unlabel A) = {}" "ground M"
  and "\<forall>(i,p) \<in> setops\<^sub>l\<^sub>s\<^sub>s\<^sub>t A \<union> set D. \<forall>(j,q) \<in> setops\<^sub>l\<^sub>s\<^sub>s\<^sub>t A \<union> set D.
        (\<exists>\<delta>. Unifier \<delta> (to_pair p) (to_pair q)) \<longrightarrow> i = j" (is "?R A D")
  and \<I>: "interpretation\<^sub>s\<^sub>u\<^sub>b\<^sub>s\<^sub>t \<I>"
  shows "\<lbrakk>M; set (unlabel D) \<cdot>\<^sub>p\<^sub>a\<^sub>i\<^sub>r\<^sub>s \<I>; unlabel A\<rbrakk>\<^sub>s \<I> \<longleftrightarrow> (\<exists>B \<in> set (tr'\<^sub>p\<^sub>c A D). \<lbrakk>M; unlabel B\<rbrakk>\<^sub>d \<I>)"
        (is "?P \<longleftrightarrow> ?Q")
proof -
  have 1: "\<forall>(t,s) \<in> set (unlabel D). (fv t \<union> fv s) \<inter> bvars\<^sub>s\<^sub>s\<^sub>t (unlabel A) = {}"
    using assms(1) unfolding unlabel_def by force

  have 2: "\<forall>(i,p) \<in> setops\<^sub>l\<^sub>s\<^sub>s\<^sub>t A \<union> set D. \<forall>(j,q) \<in> setops\<^sub>l\<^sub>s\<^sub>s\<^sub>t A \<union> set D. p = q \<longrightarrow> i = j"
    using assms(4) var_subst_id by blast

  show ?thesis by (metis tr_sem_equiv'[OF 1 assms(2,3) \<I>] tr'_par_iff_unlabel_tr[OF 2])
qed


section {* Theorem: Semantic equivalence of translation, part 4 *}
lemma tr_par_sem_equiv:
  assumes "\<forall>(l,t,s) \<in> set D. (fv t \<union> fv s) \<inter> bvars\<^sub>s\<^sub>s\<^sub>t (unlabel A) = {}"
  and "fv\<^sub>s\<^sub>s\<^sub>t (unlabel A) \<inter> bvars\<^sub>s\<^sub>s\<^sub>t (unlabel A) = {}" "ground M"
  and "\<forall>(i,p) \<in> setops\<^sub>l\<^sub>s\<^sub>s\<^sub>t A \<union> set D. \<forall>(j,q) \<in> setops\<^sub>l\<^sub>s\<^sub>s\<^sub>t A \<union> set D.
        (\<exists>\<delta>. Unifier \<delta> (to_pair p) (to_pair q)) \<longrightarrow> i = j"
  and \<I>: "interpretation\<^sub>s\<^sub>u\<^sub>b\<^sub>s\<^sub>t \<I>"
  shows "\<lbrakk>M; set (unlabel D) \<cdot>\<^sub>p\<^sub>a\<^sub>i\<^sub>r\<^sub>s \<I>; unlabel A\<rbrakk>\<^sub>s \<I> \<longleftrightarrow> (\<exists>B \<in> set (tr\<^sub>p\<^sub>c A D). \<lbrakk>M; unlabel B\<rbrakk>\<^sub>d \<I>)"
  (is "?P \<longleftrightarrow> ?Q")
using tr_par_iff_tr'_par[OF assms(4)] tr'_par_sem_equiv[OF assms] by metis


section {* Theorem: The stateful compositionality result *}
theorem par_comp_constr_stateful:
  assumes \<A>: "par_comp\<^sub>l\<^sub>s\<^sub>s\<^sub>t \<A> Sec" "typing_cond\<^sub>s\<^sub>s\<^sub>t (unlabel \<A>)"
  and \<I>: "\<I> \<Turnstile>\<^sub>s unlabel \<A>" "interpretation\<^sub>s\<^sub>u\<^sub>b\<^sub>s\<^sub>t \<I>"
  shows "\<exists>\<I>\<^sub>\<tau>. interpretation\<^sub>s\<^sub>u\<^sub>b\<^sub>s\<^sub>t \<I>\<^sub>\<tau> \<and> wt\<^sub>s\<^sub>u\<^sub>b\<^sub>s\<^sub>t \<I>\<^sub>\<tau> \<and> wf\<^sub>t\<^sub>r\<^sub>m\<^sub>s (subst_range \<I>\<^sub>\<tau>) \<and>
              ((\<forall>n. \<I>\<^sub>\<tau> \<Turnstile>\<^sub>s proj_unl n \<A>) \<or> (\<exists>\<A>'. prefix \<A>' \<A> \<and> \<A>' leaks Sec under \<I>\<^sub>\<tau>))"
proof -
  let ?P = "\<lambda>n A D.
       \<forall>(i, p) \<in> setops\<^sub>l\<^sub>s\<^sub>s\<^sub>t (proj n A) \<union> set D.
       \<forall>(j, q) \<in> setops\<^sub>l\<^sub>s\<^sub>s\<^sub>t (proj n A) \<union> set D.
          (\<exists>\<delta>. Unifier \<delta> (to_pair p) (to_pair q)) \<longrightarrow> i = j"

  have 1: "\<forall>(l, t, t')\<in>set []. (fv t \<union> fv t') \<inter> bvars\<^sub>s\<^sub>s\<^sub>t (unlabel \<A>) = {}"
          "fv\<^sub>s\<^sub>s\<^sub>t (unlabel \<A>) \<inter> bvars\<^sub>s\<^sub>s\<^sub>t (unlabel \<A>) = {}" "ground {}"
    using \<A>(2) unfolding typing_cond\<^sub>s\<^sub>s\<^sub>t_def by simp_all

  have 2: "\<And>n. \<forall>(l, t, t')\<in>set []. (fv t \<union> fv t') \<inter> bvars\<^sub>s\<^sub>s\<^sub>t (proj_unl n \<A>) = {}"
          "\<And>n. fv\<^sub>s\<^sub>s\<^sub>t (proj_unl n \<A>) \<inter> bvars\<^sub>s\<^sub>s\<^sub>t (proj_unl n \<A>) = {}"
    using 1(1,2) sst_vars_proj_subset[of _ \<A>] by fast+

  have 3: "\<And>n. par_comp\<^sub>l\<^sub>s\<^sub>s\<^sub>t (proj n \<A>) Sec"
    using par_comp\<^sub>l\<^sub>s\<^sub>s\<^sub>t_proj[OF \<A>(1)] by metis

  obtain \<A>' where \<A>': "\<A>' \<in> set (tr\<^sub>p\<^sub>c \<A> [])" "\<I> \<Turnstile> \<langle>unlabel \<A>'\<rangle>"
    using tr_par_sem_equiv[OF 1 _ \<I>(2)] \<A>(1) \<I>(1) unfolding par_comp\<^sub>l\<^sub>s\<^sub>s\<^sub>t_def by moura
  obtain \<I>\<^sub>\<tau> where \<I>\<^sub>\<tau>: 
      "interpretation\<^sub>s\<^sub>u\<^sub>b\<^sub>s\<^sub>t \<I>\<^sub>\<tau>" "wt\<^sub>s\<^sub>u\<^sub>b\<^sub>s\<^sub>t \<I>\<^sub>\<tau>" "wf\<^sub>t\<^sub>r\<^sub>m\<^sub>s (subst_range \<I>\<^sub>\<tau>)"
      "(\<forall>n. \<I>\<^sub>\<tau> \<Turnstile> \<langle>proj_unl n \<A>'\<rangle>) \<or>
       (\<exists>\<A>''. prefix \<A>'' \<A>' \<and> \<A>'' leaks Sec under \<I>\<^sub>\<tau>)"
    using par_comp_constr[OF tr_par_preserves_par_comp[OF \<A>(1) \<A>'(1)]
                             tr_par_preserves_typing_cond[OF \<A> \<A>'(1)]
                             \<A>'(2) \<I>(2)]
    by moura
  show ?thesis
  proof (cases "\<forall>n. \<I>\<^sub>\<tau> \<Turnstile> \<langle>proj_unl n \<A>'\<rangle>")
    case True
    { fix n assume "\<I>\<^sub>\<tau> \<Turnstile> \<langle>proj_unl n \<A>'\<rangle>"
      hence "\<lbrakk>{}; {}; unlabel (proj n \<A>)\<rbrakk>\<^sub>s \<I>\<^sub>\<tau>"
        using tr_par_proj[OF \<A>'(1), of n]
              tr_par_sem_equiv[OF 2(1,2) 1(3) _ \<I>\<^sub>\<tau>(1), of n] 3(1)
        unfolding par_comp\<^sub>l\<^sub>s\<^sub>s\<^sub>t_def proj_def by force
    } thus ?thesis using True \<I>\<^sub>\<tau>(1,2,3) by metis
  next
    case False
    then obtain \<A>''::"('fun,'var,'lbl) labeled_strand" where \<A>'':
        "prefix \<A>'' \<A>'" "\<A>'' leaks Sec under \<I>\<^sub>\<tau>"
      using \<I>\<^sub>\<tau> by blast
    moreover {
      fix t l assume *: "\<lbrakk>{}; unlabel (proj l \<A>'')@[send\<langle>t\<rangle>]\<rbrakk>\<^sub>d \<I>\<^sub>\<tau>"
      have "\<I>\<^sub>\<tau> \<Turnstile> \<langle>unlabel (proj l \<A>'')\<rangle>" "ik\<^sub>s\<^sub>t (unlabel (proj l \<A>'')) \<cdot>\<^sub>s\<^sub>e\<^sub>t \<I>\<^sub>\<tau> \<turnstile> t \<cdot> \<I>\<^sub>\<tau>"
        using strand_sem_split_left_d[OF *]
              strand_sem_split_right_d[OF *]
        by auto
    } ultimately have "\<exists>t\<in>Sec - declassified\<^sub>l\<^sub>s\<^sub>t \<A>'' \<I>\<^sub>\<tau>. \<exists>l.
            (\<I>\<^sub>\<tau> \<Turnstile> \<langle>unlabel (proj l \<A>'')\<rangle>) \<and> ik\<^sub>s\<^sub>t (unlabel (proj l \<A>'')) \<cdot>\<^sub>s\<^sub>e\<^sub>t \<I>\<^sub>\<tau> \<turnstile> t \<cdot> \<I>\<^sub>\<tau>"
      unfolding strand_leaks\<^sub>l\<^sub>s\<^sub>t_def constr_sem_d_def by metis
    then obtain s m where sm:
        "s \<in> Sec - declassified\<^sub>l\<^sub>s\<^sub>t \<A>'' \<I>\<^sub>\<tau>"
        "\<I>\<^sub>\<tau> \<Turnstile> \<langle>unlabel (proj m \<A>'')\<rangle>"
        "ik\<^sub>s\<^sub>t (unlabel (proj m \<A>'')) \<cdot>\<^sub>s\<^sub>e\<^sub>t \<I>\<^sub>\<tau> \<turnstile> s \<cdot> \<I>\<^sub>\<tau>"
      by moura
    (* we need to show that there is some prefix B of \<A>'' that also leaks
       and where B \<in> set (tr C D for some prefix C of \<A> *) 
    then obtain B::"('fun,'var,'lbl) labeled_strand"
        and C::"('fun,'var,'lbl) labeled_stateful_strand"
      where BC:
        "prefix B \<A>'" "prefix C \<A>" "B \<in> set (tr\<^sub>p\<^sub>c C [])"
        "ik\<^sub>s\<^sub>t (unlabel (proj m B)) \<cdot>\<^sub>s\<^sub>e\<^sub>t \<I>\<^sub>\<tau> \<turnstile> s \<cdot> \<I>\<^sub>\<tau>"
        "prefix B \<A>''"
      using tr_leaking_prefix_exists[OF \<A>'(1) \<A>''(1) sm(3)] prefix_order.order_trans[OF _ \<A>''(1)]
      by auto
    have "\<lbrakk>{}; unlabel (proj m B)\<rbrakk>\<^sub>d \<I>\<^sub>\<tau>"
      using sm(2) BC(5) unfolding prefix_def unlabel_def proj_def by auto
    hence BC': "\<I>\<^sub>\<tau> \<Turnstile> \<langle>proj_unl m B@[send\<langle>s\<rangle>]\<rangle>" using BC(4) by auto
    have BC'': "s \<in> Sec - declassified\<^sub>l\<^sub>s\<^sub>t B \<I>\<^sub>\<tau>" using BC(5) sm(1) unfolding prefix_def by auto
    have 4: "\<And>n. par_comp\<^sub>l\<^sub>s\<^sub>s\<^sub>t (proj n C) Sec"
      using \<A>(1) BC(2) par_comp\<^sub>l\<^sub>s\<^sub>s\<^sub>t_prefix[THEN par_comp\<^sub>l\<^sub>s\<^sub>s\<^sub>t_proj]
      unfolding prefix_def by auto
    have "fv\<^sub>s\<^sub>s\<^sub>t (unlabel \<A>) \<inter> bvars\<^sub>s\<^sub>s\<^sub>t (unlabel \<A>) = {}"
         "fv\<^sub>s\<^sub>s\<^sub>t (unlabel C) \<subseteq> fv\<^sub>s\<^sub>s\<^sub>t (unlabel \<A>)"
         "bvars\<^sub>s\<^sub>s\<^sub>t (unlabel C) \<subseteq> bvars\<^sub>s\<^sub>s\<^sub>t (unlabel \<A>)"
      using \<A>(2) BC(2) sst_vars_append_subset(1,2)[of "unlabel C"]
      unfolding typing_cond\<^sub>s\<^sub>s\<^sub>t_def prefix_def unlabel_def by auto
    hence "\<And>n. fv\<^sub>s\<^sub>s\<^sub>t (proj_unl n C) \<inter> bvars\<^sub>s\<^sub>s\<^sub>t (proj_unl n C) = {}"
      using sst_vars_proj_subset[of _ C] sst_vars_proj_subset[of _ \<A>]
      by blast
    hence 5:
        "\<And>n. \<forall>(l, t, t')\<in>set []. (fv t \<union> fv t') \<inter> bvars\<^sub>s\<^sub>s\<^sub>t (proj_unl n C) = {}"
        "\<And>n. fv\<^sub>s\<^sub>s\<^sub>t (proj_unl n C) \<inter> bvars\<^sub>s\<^sub>s\<^sub>t (proj_unl n C) = {}"
        "ground {}"
      using 2 by auto
    have 6: "\<And>n. ?P n C []" using 4 unfolding par_comp\<^sub>l\<^sub>s\<^sub>s\<^sub>t_def by simp
    have "s \<cdot> \<I>\<^sub>\<tau> = s" using \<I>\<^sub>\<tau>(1) BC'' \<A>(1) unfolding par_comp\<^sub>l\<^sub>s\<^sub>s\<^sub>t_def by auto
    hence "\<exists>n. (\<I>\<^sub>\<tau> \<Turnstile>\<^sub>s proj_unl n C) \<and> ik\<^sub>s\<^sub>s\<^sub>t (proj_unl n C) \<cdot>\<^sub>s\<^sub>e\<^sub>t \<I>\<^sub>\<tau> \<turnstile> s \<cdot> \<I>\<^sub>\<tau>"
      using tr_par_proj[OF BC(3), of m] BC'(1)
            tr_par_sem_equiv[OF 5 6 \<I>\<^sub>\<tau>(1), of m]
            tr_par_deduct_iff[OF tr_par_proj(1)[OF BC(3)], of \<I>\<^sub>\<tau> m s]
      unfolding proj_def by auto
    hence "\<exists>n. \<I>\<^sub>\<tau> \<Turnstile>\<^sub>s (proj_unl n C@[Send s])" using strand_sem_append_stateful by simp
    moreover have "s \<in> Sec - declassified\<^sub>l\<^sub>s\<^sub>s\<^sub>t C \<I>\<^sub>\<tau>" by (metis tr_par_declassified_eq BC(3) BC'')
    ultimately show ?thesis using \<I>\<^sub>\<tau>(1,2,3) BC(2) unfolding strand_leaks\<^sub>l\<^sub>s\<^sub>s\<^sub>t_def by metis
  qed
qed

end

end
